"""
An example for running incremental SfM on 360 spherical panorama images.
"""

import argparse
import os
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from pathlib import Path
from threading import Lock
from typing import Literal, TypeVar, cast

import cv2
import numpy as np
import numpy.typing as npt
import PIL.ExifTags
import PIL.Image
from scipy.spatial.transform import Rotation
from tqdm import tqdm

import pycolmap
from pycolmap import logging

N = TypeVar("N", bound=int)
NDArrayNx2 = np.ndarray[tuple[N, Literal[2]], np.dtype[np.float64]]
NDArray3x1 = np.ndarray[tuple[Literal[3], Literal[1]], np.dtype[np.float64]]


@dataclass
class PanoRenderOptions:
    num_steps_yaw: int
    pitches_deg: Sequence[float]
    hfov_deg: float
    vfov_deg: float


PANO_RENDER_OPTIONS: dict[str, PanoRenderOptions] = {
    "overlapping": PanoRenderOptions(
        num_steps_yaw=4,
        pitches_deg=(-35.0, 0.0, 35.0),
        hfov_deg=90.0,
        vfov_deg=90.0,
    ),
    # Cubemap without top and bottom images.
    "non-overlapping": PanoRenderOptions(
        num_steps_yaw=4,
        pitches_deg=(0.0,),
        hfov_deg=90.0,
        vfov_deg=90.0,
    ),
}


def create_virtual_camera(
    pano_width: int,
    pano_height: int,
    hfov_deg: float,
    vfov_deg: float,
) -> pycolmap.Camera:
    """Create a virtual perspective camera."""
    image_width = int(pano_width * hfov_deg / 360)
    image_height = int(pano_height * vfov_deg / 180)
    focal = image_width / (2 * np.tan(np.deg2rad(hfov_deg) / 2))
    return pycolmap.Camera.create_from_model_id(
        camera_id=0,
        model=pycolmap.CameraModelId.SIMPLE_PINHOLE,
        focal_length=focal,
        width=image_width,
        height=image_height,
    )


def get_virtual_camera_rays(
    camera: pycolmap.Camera,
) -> npt.NDArray[np.floating]:
    size = (camera.width, camera.height)
    x, y = np.indices(size).astype(np.float32)
    xy: NDArrayNx2 = np.column_stack([x.ravel(), y.ravel()])
    # The center of the upper left most pixel has coordinate (0.5, 0.5)
    xy += 0.5
    xy_norm: NDArrayNx2 = camera.cam_from_img(image_points=xy)
    rays = np.concatenate([xy_norm, np.ones_like(xy_norm[:, :1])], -1)
    rays /= np.linalg.norm(rays, axis=-1, keepdims=True)
    return rays


def spherical_img_from_cam(
    image_size: tuple[int, int], rays_in_cam: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]:
    """Project rays into a 360 panorama (spherical) image."""
    if image_size[0] != image_size[1] * 2:
        raise ValueError("Only 360° panoramas are supported.")
    if rays_in_cam.ndim != 2 or rays_in_cam.shape[1] != 3:
        raise ValueError(f"{rays_in_cam.shape=} but expected (N,3).")
    r = rays_in_cam.T
    yaw = np.arctan2(r[0], r[2])
    pitch = -np.arctan2(r[1], np.linalg.norm(r[[0, 2]], axis=0))
    u = (1 + yaw / np.pi) / 2
    v = (1 - pitch * 2 / np.pi) / 2
    return np.stack([u, v], -1) * image_size


def get_virtual_rotations(
    num_steps_yaw: int, pitches_deg: Sequence[float]
) -> Sequence[npt.NDArray[np.floating]]:
    """Get the relative rotations of the virtual cameras w.r.t. the panorama."""
    # Assuming that the panos are approximately upright.
    cams_from_pano_r = []
    yaws = np.linspace(0, 360, num_steps_yaw, endpoint=False)
    for pitch_deg in pitches_deg:
        yaw_offset = (360 / num_steps_yaw / 2) if pitch_deg > 0 else 0
        for yaw_deg in yaws + yaw_offset:
            cam_from_pano_r = Rotation.from_euler(
                "XY", [-pitch_deg, -yaw_deg], degrees=True
            ).as_matrix()
            cams_from_pano_r.append(cam_from_pano_r)
    return cams_from_pano_r


def create_pano_rig_config(
    cams_from_pano_rotation: Sequence[npt.NDArray[np.floating]],
    ref_idx: int = 0,
) -> pycolmap.RigConfig:
    """Create a RigConfig for the given virtual rotations."""
    rig_cameras = []
    zero_translation = cast(NDArray3x1, np.zeros((3, 1), dtype=np.float64))
    for idx, cam_from_pano_rotation in enumerate(cams_from_pano_rotation):
        if idx == ref_idx:
            cam_from_rig = None
        else:
            cam_from_ref_rotation = (
                cam_from_pano_rotation @ cams_from_pano_rotation[ref_idx].T
            )
            cam_from_rig = pycolmap.Rigid3d(
                pycolmap.Rotation3d(cam_from_ref_rotation),
                zero_translation,
            )
        rig_cameras.append(
            pycolmap.RigConfigCamera(
                ref_sensor=idx == ref_idx,
                image_prefix=f"pano_camera{idx}/",
                cam_from_rig=cam_from_rig,
            )
        )
    return pycolmap.RigConfig(cameras=rig_cameras)


class PanoProcessor:
    def __init__(
        self,
        pano_image_dir: Path,
        output_image_dir: Path,
        mask_dir: Path,
        render_options: PanoRenderOptions,
    ):
        self.render_options = render_options
        self.pano_image_dir = pano_image_dir
        self.output_image_dir = output_image_dir
        self.mask_dir = mask_dir

        self.cams_from_pano_rotation = get_virtual_rotations(
            num_steps_yaw=render_options.num_steps_yaw,
            pitches_deg=render_options.pitches_deg,
        )
        self.rig_config = create_pano_rig_config(self.cams_from_pano_rotation)

        # We assign each pano pixel to the virtual camera
        # with the closest camera center.
        self.cam_centers_in_pano = np.einsum(
            "nij,i->nj", self.cams_from_pano_rotation, [0, 0, 1]
        )

        self._lock = Lock()

        # These are initialized on the first pano image
        # to avoid recomputing the rays for each pano image.
        self._camera: pycolmap.Camera | None = None
        self._pano_size: tuple[int, int] | None = None
        self._rays_in_cam: npt.NDArray[np.floating] | None = None

    def process(self, pano_name: str) -> None:
        pano_path = self.pano_image_dir / pano_name
        try:
            pano_pil_image = PIL.Image.open(pano_path)
        except PIL.Image.UnidentifiedImageError:
            logging.info(f"Skipping file {pano_path} as it cannot be read.")
            return

        pano_exif = pano_pil_image.getexif()
        pano_image = np.asarray(pano_pil_image)
        gpsonly_exif = PIL.Image.Exif()
        gpsonly_exif[PIL.ExifTags.IFD.GPSInfo] = pano_exif.get_ifd(
            PIL.ExifTags.IFD.GPSInfo
        )

        pano_height, pano_width, *_ = pano_image.shape
        if pano_width != pano_height * 2:
            raise ValueError("Only 360° panoramas are supported.")

        with self._lock:
            if self._camera is None:  # First image, precompute rays once.
                self._camera = create_virtual_camera(
                    pano_width=pano_width,
                    pano_height=pano_height,
                    hfov_deg=self.render_options.hfov_deg,
                    vfov_deg=self.render_options.vfov_deg,
                )
                for rig_camera in self.rig_config.cameras:
                    rig_camera.camera = self._camera
                self._pano_size = (pano_width, pano_height)
                self._rays_in_cam = get_virtual_camera_rays(self._camera)
            else:  # Later images, verify consistent panoramas.
                if (pano_width, pano_height) != self._pano_size:
                    raise ValueError(
                        "Panoramas of different sizes are not supported."
                    )

        for cam_idx, cam_from_pano_r in enumerate(self.cams_from_pano_rotation):
            assert self._rays_in_cam is not None
            rays_in_pano = self._rays_in_cam @ cam_from_pano_r
            xy_in_pano = spherical_img_from_cam(self._pano_size, rays_in_pano)
            xy_in_pano = xy_in_pano.reshape(
                self._camera.width, self._camera.height, 2
            ).astype(np.float32)
            xy_in_pano -= 0.5  # COLMAP to OpenCV pixel origin.
            x_coords, y_coords = np.moveaxis(xy_in_pano, [0, 1, 2], [2, 1, 0])
            image = cv2.remap(
                pano_image,
                x_coords,
                y_coords,
                cv2.INTER_LINEAR,
                borderMode=cv2.BORDER_WRAP,
            )
            # We define a mask such that each pixel of the panorama has its
            # features extracted only in a single virtual camera.
            closest_camera = np.argmax(
                rays_in_pano @ self.cam_centers_in_pano.T, -1
            )
            mask = (
                ((closest_camera == cam_idx) * 255)
                .astype(np.uint8)
                .reshape(self._camera.width, self._camera.height)
                .transpose()
            )

            image_name = (
                self.rig_config.cameras[cam_idx].image_prefix + pano_name
            )
            mask_name = f"{image_name}.png"

            image_path = self.output_image_dir / image_name
            image_path.parent.mkdir(exist_ok=True, parents=True)
            PIL.Image.fromarray(image).save(image_path, exif=gpsonly_exif)

            mask_path = self.mask_dir / mask_name
            mask_path.parent.mkdir(exist_ok=True, parents=True)
            if not pycolmap.Bitmap.from_array(mask).write(mask_path):
                raise RuntimeError(f"Cannot write {mask_path}")


def render_perspective_images(
    pano_image_names: Sequence[str],
    pano_image_dir: Path,
    output_image_dir: Path,
    mask_dir: Path,
    render_options: PanoRenderOptions,
) -> pycolmap.RigConfig:
    processor = PanoProcessor(
        pano_image_dir, output_image_dir, mask_dir, render_options
    )

    num_panos = len(pano_image_names)
    max_workers = min(32, (os.cpu_count() or 2) - 1)

    with tqdm(total=num_panos) as pbar:
        with ThreadPoolExecutor(max_workers=max_workers) as thread_pool:
            futures = [
                thread_pool.submit(processor.process, pano_name)
                for pano_name in pano_image_names
            ]
            for future in as_completed(futures):
                future.result()
                pbar.update(1)

    return processor.rig_config


def run(args: argparse.Namespace) -> None:
    pycolmap.set_random_seed(0)

    # Define the paths.
    image_dir = args.output_path / "images"
    mask_dir = args.output_path / "masks"
    image_dir.mkdir(exist_ok=True, parents=True)
    mask_dir.mkdir(exist_ok=True, parents=True)

    database_path = args.output_path / "database.db"
    if database_path.exists():
        database_path.unlink()

    rec_path = args.output_path / "sparse"
    rec_path.mkdir(exist_ok=True, parents=True)

    # Search for input images.
    pano_image_dir = args.input_image_path
    pano_image_names = sorted(
        p.relative_to(pano_image_dir).as_posix()
        for p in pano_image_dir.rglob("*")
        if not p.is_dir()
    )
    logging.info(f"Found {len(pano_image_names)} images in {pano_image_dir}.")

    rig_config = render_perspective_images(
        pano_image_names,
        pano_image_dir,
        image_dir,
        mask_dir,
        PANO_RENDER_OPTIONS[args.pano_render_type],
    )

    pycolmap.extract_features(
        database_path,
        image_dir,
        reader_options=pycolmap.ImageReaderOptions(mask_path=mask_dir),
        camera_mode=pycolmap.CameraMode.PER_FOLDER,
    )

    with pycolmap.Database.open(database_path) as db:
        pycolmap.apply_rig_config([rig_config], db)

    matching_options = pycolmap.FeatureMatchingOptions()
    # We have perfect sensor_from_rig poses (except for potential stitching
    # artifacts by the spherical image provider), so we can perform geometric
    # verification using rig constraints.
    matching_options.rig_verification = True
    # The images within a frame do not have overlap due to the provided masks.
    matching_options.skip_image_pairs_in_same_frame = True
    if args.matcher == "sequential":
        pycolmap.match_sequential(
            database_path,
            pairing_options=pycolmap.SequentialPairingOptions(
                loop_detection=True
            ),
            matching_options=matching_options,
        )
    elif args.matcher == "exhaustive":
        pycolmap.match_exhaustive(
            database_path, matching_options=matching_options
        )
    elif args.matcher == "vocabtree":
        pycolmap.match_vocabtree(
            database_path, matching_options=matching_options
        )
    elif args.matcher == "spatial":
        pycolmap.match_spatial(database_path, matching_options=matching_options)
    else:
        logging.fatal(f"Unknown matcher: {args.matcher}")

    opts = pycolmap.IncrementalPipelineOptions(
        ba_refine_sensor_from_rig=False,
        ba_refine_focal_length=False,
        ba_refine_principal_point=False,
        ba_refine_extra_params=False,
    )
    recs = pycolmap.incremental_mapping(
        database_path, image_dir, rec_path, opts
    )
    for idx, rec in recs.items():
        logging.info(f"#{idx} {rec.summary()}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_image_path", type=Path, required=True)
    parser.add_argument("--output_path", type=Path, required=True)
    parser.add_argument(
        "--matcher",
        default="sequential",
        choices=["sequential", "exhaustive", "vocabtree", "spatial"],
    )
    parser.add_argument(
        "--pano_render_type",
        default="overlapping",
        choices=list(PANO_RENDER_OPTIONS.keys()),
    )
    run(parser.parse_args())



"""
Runs the COLMAP automatic reconstruction pipeline on the ETH3D dataset
and asserts that the reconstructed model is close to the ground truth.

This script is intended to be run as a CI test. It is not intended to be
run manually. Instead use benchmark/reconstruction/evaluate.py.

"""

import argparse
import logging
import os
import subprocess
import sys
import urllib.request


def download_file(url: str, file_path: str, max_retries: int = 3) -> None:
    if os.path.exists(file_path):
        return
    logging.info(f"Downloading {url} to {file_path}")
    for retry in range(max_retries):
        try:
            urllib.request.urlretrieve(url, file_path)
            return
        except Exception as exc:
            logging.error(
                f"Failed to download {url} (trial={retry + 1}) "
                f"to {file_path} due to {exc}"
            )


def check_small_errors_or_exit(
    dataset_name: str,
    max_rotation_error: float,
    max_proj_center_error: float,
    expected_num_images: float,
    errors_csv_path: str,
) -> None:
    logging.info(f"Evaluating errors for {dataset_name}")

    error = False
    with open(errors_csv_path) as fid:
        num_images = 0
        for line in fid:
            line = line.strip()
            if len(line) == 0 or line.startswith("#"):
                continue
            rotation_error, proj_center_error = map(float, line.split(","))
            num_images += 1
            if rotation_error > max_rotation_error:
                logging.info(
                    "Exceeded rotation error threshold:", rotation_error
                )
                error = True
            if proj_center_error > max_proj_center_error:
                logging.info(
                    "Exceeded projection center error threshold:",
                    proj_center_error,
                )
                error = True

    if num_images != expected_num_images:
        logging.error("Unexpected number of images:", num_images)
        error = True

    if error:
        sys.exit(1)


def process_dataset(args: argparse.Namespace, dataset_name: str) -> None:
    logging.info("Processing dataset:", dataset_name)

    workspace_path = os.path.join(
        os.path.realpath(args.workspace_path), dataset_name
    )
    os.makedirs(workspace_path, exist_ok=True)

    dataset_archive_path = os.path.join(workspace_path, f"{dataset_name}.7z")
    download_file(
        f"https://www.eth3d.net/data/{dataset_name}_dslr_undistorted.7z",
        dataset_archive_path,
    )

    subprocess.check_call(
        ["7zz", "x", "-y", f"{dataset_name}.7z"], cwd=workspace_path
    )

    # Find undistorted parameters of first camera and
    # initialize all images with it. This is an approximation
    # because not all datasets have only a single camera.
    # However, it is a good enough initialization.
    with open(
        os.path.join(
            workspace_path,
            f"{dataset_name}/dslr_calibration_undistorted/cameras.txt",
        ),
    ) as fid:
        for line in fid:
            if not line.startswith("#"):
                first_camera_data = line.split()
                camera_model = first_camera_data[1]
                assert camera_model == "PINHOLE"
                camera_params = first_camera_data[4:]
                assert len(camera_params) == 4
                break

    # Count the number of expected images in the GT.
    expected_num_images = 0
    with open(
        os.path.join(
            workspace_path,
            f"{dataset_name}/dslr_calibration_undistorted/images.txt",
        ),
    ) as fid:
        for line in fid:
            if not line.startswith("#") and line.strip():
                expected_num_images += 1
    # Each image uses two consecutive lines.
    assert expected_num_images % 2 == 0
    expected_num_images //= 2

    # Run automatic reconstruction pipeline.
    subprocess.check_call(
        [
            os.path.realpath(args.colmap_path),
            "automatic_reconstructor",
            "--image_path",
            f"{dataset_name}/images/",
            "--workspace_path",
            workspace_path,
            "--use_gpu",
            "1" if args.use_gpu else "0",
            "--num_threads",
            str(args.num_threads),
            "--quality",
            args.quality,
            "--camera_model",
            "PINHOLE",
            "--camera_params",
            ",".join(camera_params),
        ],
        cwd=workspace_path,
    )

    # Compare reconstructed model to GT model.
    subprocess.check_call(
        [
            os.path.realpath(args.colmap_path),
            "model_comparer",
            "--input_path1",
            "sparse/0",
            "--input_path2",
            f"{dataset_name}/dslr_calibration_undistorted/",
            "--output_path",
            ".",
            "--alignment_error",
            "proj_center",
            "--max_proj_center_error",
            str(args.max_proj_center_error),
        ],
        cwd=workspace_path,
    )

    # Ensure discrepancy between reconstructed model and GT is small.
    check_small_errors_or_exit(
        dataset_name,
        args.max_rotation_error,
        args.max_proj_center_error,
        expected_num_images,
        os.path.join(workspace_path, "errors.csv"),
    )


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset_names", nargs="+", required=True)
    parser.add_argument("--workspace_path", required=True)
    parser.add_argument("--colmap_path", required=True)
    parser.add_argument("--use_gpu", default=True, action="store_true")
    parser.add_argument("--use_cpu", dest="use_gpu", action="store_false")
    parser.add_argument("--num_threads", type=int, default=-1)
    parser.add_argument("--quality", default="medium")
    parser.add_argument("--max_rotation_error", type=float, default=1.0)
    parser.add_argument("--max_proj_center_error", type=float, default=0.1)
    return parser.parse_args()


def main() -> None:
    args = parse_args()

    for dataset_name in args.dataset_names:
        process_dataset(args, dataset_name)


if __name__ == "__main__":
    main()



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse

import numpy as np
import numpy.typing as npt
import open3d

import pycolmap


class Model:
    def __init__(self) -> None:
        self.reconstruction: pycolmap.Reconstruction
        self.visualizer: open3d.visualization.Visualizer

    def read_model(self, path: str) -> None:
        self.reconstruction = pycolmap.Reconstruction(path)

    def add_points(
        self, min_track_len: int = 3, remove_statistical_outlier: bool = True
    ) -> None:
        pcd = open3d.geometry.PointCloud()

        xyz = []
        rgb = []
        for point in self.reconstruction.points3D.values():
            if point.track.length() < min_track_len:
                continue
            xyz.append(point.xyz)
            rgb.append(point.color / 255)

        pcd.points = open3d.utility.Vector3dVector(xyz)
        pcd.colors = open3d.utility.Vector3dVector(rgb)

        # remove obvious outliers
        if remove_statistical_outlier:
            [pcd, _] = pcd.remove_statistical_outlier(
                nb_neighbors=20, std_ratio=2.0
            )

        # open3d.visualization.draw_geometries([pcd])
        self.visualizer.add_geometry(pcd)
        self.visualizer.poll_events()
        self.visualizer.update_renderer()

    def add_cameras(self, scale: float = 1) -> None:
        frustums = []
        for img in self.reconstruction.images.values():
            # extrinsics
            world_from_cam = img.cam_from_world().inverse()
            R = world_from_cam.rotation.matrix()
            t = world_from_cam.translation

            # intrinsics
            cam = img.camera
            if cam.model in (
                pycolmap.CameraModelId.SIMPLE_PINHOLE,
                pycolmap.CameraModelId.SIMPLE_RADIAL,
                pycolmap.CameraModelId.RADIAL,
            ):
                fx = fy = cam.params[0]
                cx = cam.params[1]
                cy = cam.params[2]
            elif cam.model in (
                pycolmap.CameraModelId.PINHOLE,
                pycolmap.CameraModelId.OPENCV,
                pycolmap.CameraModelId.OPENCV_FISHEYE,
                pycolmap.CameraModelId.FULL_OPENCV,
            ):
                fx = cam.params[0]
                fy = cam.params[1]
                cx = cam.params[2]
                cy = cam.params[3]
            else:
                raise Exception("Camera model not supported")

            # intrinsics
            K = np.identity(3)
            K[0, 0] = fx
            K[1, 1] = fy
            K[0, 2] = cx
            K[1, 2] = cy

            # create axis, plane and pyramid geometries that will be drawn
            cam_model = draw_camera(K, R, t, cam.width, cam.height, scale)
            frustums.extend(cam_model)

        # add geometries to visualizer
        for i in frustums:
            self.visualizer.add_geometry(i)

    def create_window(self) -> None:
        self.visualizer = open3d.visualization.Visualizer()
        self.visualizer.create_window()

    def show(self) -> None:
        self.visualizer.poll_events()
        self.visualizer.update_renderer()
        self.visualizer.run()
        self.visualizer.destroy_window()


def draw_camera(
    K: npt.NDArray[np.float64],
    R: npt.NDArray[np.float64],
    t: npt.NDArray[np.float64],
    w: int,
    h: int,
    scale: float = 1,
    color: list[float] | None = None,
) -> list[open3d.geometry.Geometry]:
    """Create axis, plane and pyramed geometries in Open3D format.
    :param K: calibration matrix (camera intrinsics)
    :param R: rotation matrix
    :param t: translation
    :param w: image width
    :param h: image height
    :param scale: camera model scale
    :param color: color of the image plane and pyramid lines
    :return: camera model geometries (axis, plane and pyramid)
    """

    if color is None:
        color = [0.8, 0.2, 0.8]

    # intrinsics
    K = K.copy() / scale
    Kinv = np.linalg.inv(K)

    # 4x4 transformation
    T = np.column_stack((R, t))
    T = np.vstack((T, (0, 0, 0, 1)))

    # axis
    axis = open3d.geometry.TriangleMesh.create_coordinate_frame(
        size=0.5 * scale
    )
    axis.transform(T)

    # points in pixel
    points_pixel = [
        [0, 0, 0],
        [0, 0, 1],
        [w, 0, 1],
        [0, h, 1],
        [w, h, 1],
    ]

    # pixel to camera coordinate system
    points = [Kinv @ p for p in points_pixel]

    # image plane
    width = abs(points[1][0]) + abs(points[3][0])
    height = abs(points[1][1]) + abs(points[3][1])
    plane = open3d.geometry.TriangleMesh.create_box(width, height, depth=1e-6)
    plane.paint_uniform_color(color)
    plane.translate([points[1][0], points[1][1], scale])
    plane.transform(T)

    # pyramid
    points_in_world = [(R @ p + t) for p in points]
    lines = [
        [0, 1],
        [0, 2],
        [0, 3],
        [0, 4],
    ]
    colors = [color for i in range(len(lines))]
    line_set = open3d.geometry.LineSet(
        points=open3d.utility.Vector3dVector(points_in_world),
        lines=open3d.utility.Vector2iVector(lines),
    )
    line_set.colors = open3d.utility.Vector3dVector(colors)

    # return as list in Open3D format
    return [axis, plane, line_set]


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Visualize COLMAP binary and text models"
    )
    parser.add_argument(
        "--input_model", required=True, help="path to input model folder"
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()

    # read COLMAP model
    model = Model()
    model.read_model(args.input_model)

    print("num_cameras:", model.reconstruction.num_cameras())
    print("num_images:", model.reconstruction.num_images())
    print("num_points3D:", model.reconstruction.num_points3D())

    # display using Open3D visualization tools
    model.create_window()
    model.add_points()
    model.add_cameras(scale=0.25)
    model.show()


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from .. import _core
from ..utils import import_module_symbols

if TYPE_CHECKING:
    from .._core.pyceres import *  # noqa

__all__ = import_module_symbols(globals(), _core.pyceres)
del _core



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import pytest

import pycolmap

from .utils import (
    Metrics,
    compute_abs_errors,
    compute_auc,
    compute_avg_metrics,
    compute_recall,
    compute_rel_errors,
    diff_metrics,
    get_scores,
    normalize_vec,
    vec_angular_dist_deg,
)


class TestNormalizeVec:
    def test_unit_vector(self):
        vec = np.array([1.0, 0.0, 0.0])
        normalized = normalize_vec(vec)
        np.testing.assert_allclose(normalized, vec)
        np.testing.assert_almost_equal(np.linalg.norm(normalized), 1.0)

    def test_non_unit_vector(self):
        vec = np.array([3.0, 4.0, 0.0])
        normalized = normalize_vec(vec)
        expected = np.array([0.6, 0.8, 0.0])
        np.testing.assert_allclose(normalized, expected)
        np.testing.assert_almost_equal(np.linalg.norm(normalized), 1.0)

    def test_zero_vector(self):
        vec = np.array([0.0, 0.0, 0.0])
        normalized = normalize_vec(vec)
        assert np.linalg.norm(normalized) < 1e-8


class TestVecAngularDistDeg:
    def test_identical_vectors(self):
        vec1 = np.array([1.0, 0.0, 0.0])
        vec2 = np.array([1.0, 0.0, 0.0])
        dist = vec_angular_dist_deg(vec1, vec2)
        np.testing.assert_almost_equal(dist, 0.0)

    def test_opposite_vectors(self):
        vec1 = np.array([1.0, 0.0, 0.0])
        vec2 = np.array([-1.0, 0.0, 0.0])
        dist = vec_angular_dist_deg(vec1, vec2)
        np.testing.assert_almost_equal(dist, 180.0)

    def test_perpendicular_vectors(self):
        vec1 = np.array([1.0, 0.0, 0.0])
        vec2 = np.array([0.0, 1.0, 0.0])
        dist = vec_angular_dist_deg(vec1, vec2)
        np.testing.assert_almost_equal(dist, 90.0)

    def test_45_degree_vectors(self):
        vec1 = np.array([1.0, 0.0, 0.0])
        vec2 = np.array([1.0, 1.0, 0.0])
        dist = vec_angular_dist_deg(vec1, vec2)
        np.testing.assert_almost_equal(dist, 45.0)

    def test_non_unit_vectors(self):
        vec1 = np.array([2.0, 0.0, 0.0])
        vec2 = np.array([0.0, 3.0, 0.0])
        dist = vec_angular_dist_deg(vec1, vec2)
        # Should normalize internally
        np.testing.assert_almost_equal(dist, 90.0)

    def test_clipping_behavior(self):
        # Test that dot product is clipped to [-1, 1] to avoid numerical issues
        vec1 = np.array([1.0, 1e-10, 1e-10])
        vec2 = np.array([1.0, -1e-10, -1e-10])
        dist = vec_angular_dist_deg(vec1, vec2)
        # Should not raise error even with potential numerical issues
        assert 0 <= dist <= 180


class TestComputeAuc:
    def test_simple_uniform_errors(self):
        errors = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
        thresholds = np.array([0.25, 0.5, 1.0])
        aucs = compute_auc(errors, thresholds)
        np.testing.assert_almost_equal(aucs[0], 24.0, decimal=5)
        np.testing.assert_almost_equal(aucs[1], 50.0, decimal=5)
        np.testing.assert_almost_equal(aucs[2], 75.0, decimal=5)

    def test_all_errors_zero(self):
        errors = np.array([0.0, 0.0, 0.0])
        thresholds = np.array([0.5, 1.0])
        aucs = compute_auc(errors, thresholds)
        np.testing.assert_array_almost_equal(aucs, [100.0, 100.0])

    def test_empty_errors(self):
        errors = np.array([])
        thresholds = np.array([0.5, 1.0])
        with pytest.raises(ValueError, match="No errors to evaluate"):
            compute_auc(errors, thresholds)

    def test_all_errors_above_threshold(self):
        errors = np.array([10.0, 20.0, 30.0])
        thresholds = np.array([5.0])
        aucs = compute_auc(errors, thresholds)
        np.testing.assert_almost_equal(aucs[0], 0.0)

    def test_all_errors_below_threshold(self):
        errors = np.array([0.1, 0.2, 0.3])
        thresholds = np.array([1.0])
        aucs = compute_auc(errors, thresholds)
        np.testing.assert_almost_equal(aucs[0], 85.0, decimal=5)

    def test_inf_errors(self):
        errors = np.array([0.1, 0.2, np.inf, np.inf])
        thresholds = np.array([0.5, 1.0])
        aucs = compute_auc(errors, thresholds)
        assert np.all(aucs >= 0)
        assert np.all(aucs <= 100)

    def test_single_error(self):
        errors = np.array([0.5])
        thresholds = np.array([0.3, 1.0])
        aucs = compute_auc(errors, thresholds)
        assert len(aucs) == 2
        np.testing.assert_almost_equal(aucs[0], 0.0)
        np.testing.assert_almost_equal(aucs[1], 75.0)


class TestComputeRecall:
    def test_basic_recall(self):
        errors = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
        thresholds = np.array([0.05, 0.25, 0.5, 1.0])
        recalls = compute_recall(errors, thresholds)
        assert len(recalls) == 4
        assert recalls[3] >= recalls[2] >= recalls[1] >= recalls[0]
        assert np.all(recalls >= 0)
        assert np.all(recalls <= 100)

    def test_empty_errors(self):
        errors = np.array([])
        thresholds = np.array([0.5, 1.0])
        with pytest.raises(ValueError, match="No errors to evaluate"):
            compute_recall(errors, thresholds)

    def test_all_errors_above_threshold(self):
        errors = np.array([10.0, 20.0, 30.0])
        thresholds = np.array([5.0])
        recalls = compute_recall(errors, thresholds)
        np.testing.assert_almost_equal(recalls[0], 0.0)

    def test_all_errors_below_threshold(self):
        errors = np.array([0.1, 0.2, 0.3])
        thresholds = np.array([1.0])
        recalls = compute_recall(errors, thresholds)
        np.testing.assert_almost_equal(recalls[0], 100.0)

    def test_exact_threshold(self):
        errors = np.array([0.1, 0.5, 0.9])
        thresholds = np.array([0.5])
        recalls = compute_recall(errors, thresholds)
        np.testing.assert_almost_equal(recalls[0], 200.0 / 3.0)

    def test_multiple_thresholds(self):
        errors = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        thresholds = np.array([2.0, 3.0, 4.0])
        recalls = compute_recall(errors, thresholds)
        np.testing.assert_almost_equal(recalls[0], 40.0)
        np.testing.assert_almost_equal(recalls[1], 60.0)
        np.testing.assert_almost_equal(recalls[2], 80.0)


class TestComputeAvgMetrics:
    def test_single_scene(self):
        scene_metrics = {
            "scene1": Metrics(
                aucs=np.array([10.0, 20.0, 30.0]),
                recalls=np.array([15.0, 25.0, 35.0]),
                error_thresholds=np.array([0.5, 1.0, 2.0]),
                error_type="relative_auc",
                num_images=100,
                num_reg_images=90,
                num_components=1,
                largest_component=90,
            )
        }
        aucs, recalls = compute_avg_metrics(scene_metrics)
        np.testing.assert_array_equal(aucs, [10.0, 20.0, 30.0])
        np.testing.assert_array_equal(recalls, [15.0, 25.0, 35.0])

    def test_multiple_scenes(self):
        scene_metrics = {
            "scene1": Metrics(
                aucs=np.array([10.0, 20.0, 30.0]),
                recalls=np.array([15.0, 25.0, 35.0]),
                error_thresholds=np.array([0.5, 1.0, 2.0]),
                error_type="relative_auc",
                num_images=100,
                num_reg_images=90,
                num_components=1,
                largest_component=90,
            ),
            "scene2": Metrics(
                aucs=np.array([20.0, 30.0, 40.0]),
                recalls=np.array([25.0, 35.0, 45.0]),
                error_thresholds=np.array([0.5, 1.0, 2.0]),
                error_type="relative_auc",
                num_images=100,
                num_reg_images=90,
                num_components=1,
                largest_component=90,
            ),
        }
        aucs, recalls = compute_avg_metrics(scene_metrics)
        np.testing.assert_array_equal(aucs, [15.0, 25.0, 35.0])
        np.testing.assert_array_equal(recalls, [20.0, 30.0, 40.0])

    def test_skip_special_scenes(self):
        scene_metrics = {
            "scene1": Metrics(
                aucs=np.array([10.0, 20.0, 30.0]),
                recalls=np.array([15.0, 25.0, 35.0]),
                error_thresholds=np.array([0.5, 1.0, 2.0]),
                error_type="relative_auc",
                num_images=100,
                num_reg_images=90,
                num_components=1,
                largest_component=90,
            ),
            "__avg__": Metrics(
                aucs=np.array([50.0, 60.0, 70.0]),
                recalls=np.array([55.0, 65.0, 75.0]),
                error_thresholds=np.array([0.5, 1.0, 2.0]),
                error_type="relative_auc",
                num_images=100,
                num_reg_images=90,
                num_components=1,
                largest_component=90,
            ),
            "__all__": Metrics(
                aucs=np.array([80.0, 90.0, 100.0]),
                recalls=np.array([85.0, 95.0, 105.0]),
                error_thresholds=np.array([0.5, 1.0, 2.0]),
                error_type="relative_auc",
                num_images=100,
                num_reg_images=90,
                num_components=1,
                largest_component=90,
            ),
        }
        aucs, recalls = compute_avg_metrics(scene_metrics)
        # Should only average scene1, not __avg__ or __all__
        np.testing.assert_array_equal(aucs, [10.0, 20.0, 30.0])
        np.testing.assert_array_equal(recalls, [15.0, 25.0, 35.0])


class TestGetScores:
    def test_get_auc_scores(self):
        metrics = Metrics(
            aucs=np.array([10.0, 20.0, 30.0]),
            recalls=np.array([15.0, 25.0, 35.0]),
            error_thresholds=np.array([0.5, 1.0, 2.0]),
            error_type="relative_auc",
            num_images=100,
            num_reg_images=90,
            num_components=1,
            largest_component=90,
        )
        scores = get_scores("relative_auc", metrics)
        np.testing.assert_array_equal(scores, metrics.aucs)

    def test_get_recall_scores(self):
        metrics = Metrics(
            aucs=np.array([10.0, 20.0, 30.0]),
            recalls=np.array([15.0, 25.0, 35.0]),
            error_thresholds=np.array([0.5, 1.0, 2.0]),
            error_type="relative_recall",
            num_images=100,
            num_reg_images=90,
            num_components=1,
            largest_component=90,
        )
        scores = get_scores("relative_recall", metrics)
        np.testing.assert_array_equal(scores, metrics.recalls)


class TestDiffMetrics:
    def test_nominal(self):
        metrics_a = {
            "dataset1": {
                "category1": {
                    "scene1": Metrics(
                        aucs=np.array([20.0, 30.0, 40.0]),
                        recalls=np.array([25.0, 35.0, 45.0]),
                        error_thresholds=np.array([0.5, 1.0, 2.0]),
                        error_type="relative_auc",
                        num_images=100,
                        num_reg_images=90,
                        num_components=2,
                        largest_component=80,
                    )
                }
            }
        }
        metrics_b = {
            "dataset1": {
                "category1": {
                    "scene1": Metrics(
                        aucs=np.array([10.0, 20.0, 30.0]),
                        recalls=np.array([15.0, 25.0, 35.0]),
                        error_thresholds=np.array([0.5, 1.0, 2.0]),
                        error_type="relative_auc",
                        num_images=100,
                        num_reg_images=85,
                        num_components=1,
                        largest_component=85,
                    )
                }
            }
        }
        diff = diff_metrics(metrics_a, metrics_b)

        scene_diff = diff["dataset1"]["category1"]["scene1"]
        np.testing.assert_array_equal(scene_diff.aucs, [10.0, 10.0, 10.0])
        np.testing.assert_array_equal(scene_diff.recalls, [10.0, 10.0, 10.0])
        assert scene_diff.num_reg_images == 5
        assert scene_diff.num_components == 1


def create_test_reconstruction():
    pycolmap.set_random_seed(0)
    synthetic_dataset_options = pycolmap.SyntheticDatasetOptions()
    synthetic_dataset_options.num_cameras_per_rig = 1
    synthetic_dataset_options.num_frames_per_rig = 5
    synthetic_dataset_options.num_points3D = 0
    return pycolmap.synthesize_dataset(synthetic_dataset_options)


class TestComputeAbsErrors:
    def test_identical_reconstruction(self):
        reconstruction = create_test_reconstruction()

        dts, dRs = compute_abs_errors(
            sparse_gt=reconstruction, sparse=reconstruction
        )

        assert len(dts) == reconstruction.num_images()
        assert len(dRs) == reconstruction.num_images()
        np.testing.assert_allclose(dts, 0.0, atol=1e-10)
        np.testing.assert_allclose(dRs, 0.0, atol=1e-10)

    def test_transformed_reconstruction(self):
        gt_reconstruction = create_test_reconstruction()
        reconstruction = create_test_reconstruction()
        translation = np.array([1, 2, 3])
        for frame in reconstruction.frames.values():
            world_from_rig = frame.rig_from_world.inverse()
            world_from_rig.rotation = (
                world_from_rig.rotation * pycolmap.Rotation3d([0, 1, 0, 0])
            )
            world_from_rig.translation += translation
            frame.rig_from_world = world_from_rig.inverse()

        dts, dRs = compute_abs_errors(
            sparse_gt=gt_reconstruction, sparse=reconstruction
        )

        assert len(dts) == reconstruction.num_images()
        assert len(dRs) == reconstruction.num_images()
        np.testing.assert_allclose(dts, np.linalg.norm(translation), atol=1e-10)
        np.testing.assert_allclose(dRs, 180.0, atol=1e-10)


class TestComputeRelErrors:
    def test_identical_reconstruction(self):
        reconstruction = create_test_reconstruction()

        dts, dRs = compute_rel_errors(
            sparse_gt=reconstruction,
            sparse=reconstruction,
            min_proj_center_dist=0.01,
        )

        num_images = reconstruction.num_images()
        expected_num_errors = num_images * (num_images - 1)
        assert len(dts) == expected_num_errors
        assert len(dRs) == expected_num_errors
        np.testing.assert_allclose(dts, 0.0, atol=1e-5)
        np.testing.assert_allclose(dRs, 0.0, atol=1e-5)

    def test_transformed_reconstruction(self):
        gt_reconstruction = create_test_reconstruction()
        reconstruction = create_test_reconstruction()
        reconstruction.transform(
            pycolmap.Sim3d(
                1.0,
                pycolmap.Rotation3d(np.array([0, 1, 0, 0])),
                np.array([1, 2, 3]),
            )
        )

        dts, dRs = compute_rel_errors(
            sparse_gt=gt_reconstruction,
            sparse=reconstruction,
            min_proj_center_dist=0.01,
        )

        num_images = reconstruction.num_images()
        expected_num_errors = num_images * (num_images - 1)
        assert len(dts) == expected_num_errors
        assert len(dRs) == expected_num_errors
        np.testing.assert_allclose(dts, 0.0, atol=1e-5)
        np.testing.assert_allclose(dRs, 0.0, atol=1e-5)

    def test_different_reconstructions(self):
        gt_reconstruction = create_test_reconstruction()
        reconstruction = create_test_reconstruction()
        for image in reconstruction.images.values():
            image.frame.rig_from_world.rotation = (
                pycolmap.Rotation3d(np.array([0, 1, 0, 0]))
                * image.frame.rig_from_world.rotation
            )
            image.frame.rig_from_world.translation += np.array([1, 2, 3])

        dts, dRs = compute_rel_errors(
            sparse_gt=gt_reconstruction,
            sparse=reconstruction,
            min_proj_center_dist=0.01,
        )

        num_images = reconstruction.num_images()
        expected_num_errors = num_images * (num_images - 1)
        assert len(dts) == expected_num_errors
        assert len(dRs) == expected_num_errors
        assert np.all(dts > 0.1)
        assert np.all(dRs > 0.1)



from collections.abc import MutableSequence
from types import ModuleType
from typing import Any


def import_module_symbols(
    dst_vars: dict[str, Any],
    src_module: ModuleType,
    exclude: set[str] | None = None,
) -> MutableSequence[str]:
    symbols = {}
    for n, s in vars(src_module).items():
        if n.startswith("_"):
            continue
        if exclude is not None and n in exclude:
            continue
        symbols[n] = s
    dst_vars.update(symbols)
    return list(symbols)



"""
Python reimplementation of the C++ incremental mapper with equivalent logic.
"""

import argparse
import time
from pathlib import Path

import custom_bundle_adjustment
import enlighten

import pycolmap
from pycolmap import (
    IncrementalMapper,
    IncrementalMapperOptions,
    IncrementalPipeline,
    IncrementalPipelineCallback,
    IncrementalPipelineOptions,
    IncrementalPipelineStatus,
    Reconstruction,
    ReconstructionManager,
    logging,
)


def write_snapshot(reconstruction: Reconstruction, snapshot_path: Path) -> None:
    logging.info("Creating snapshot")
    timestamp = time.time() * 1000
    path = snapshot_path / f"{timestamp:010d}"
    path.mkdir(exist_ok=True, parents=True)
    logging.verbose(1, f"=> Writing to {path}")
    reconstruction.write(path)


def has_unknown_sensor_from_rig(
    reconstruction: Reconstruction,
) -> bool:
    parameterized_rig_ids = set()
    for image in reconstruction.images.values():
        parameterized_rig_ids.add(image.frame.rig_id)
    for rig_id in parameterized_rig_ids:
        rig = reconstruction.rig(rig_id)
        for sensor_id, sensor_from_rig in rig.non_ref_sensors.items():
            if (
                sensor_id.type == pycolmap.SensorType.CAMERA
                and sensor_from_rig is None
            ):
                return True
    return False


def iterative_global_refinement(
    options: IncrementalPipelineOptions,
    mapper_options: IncrementalMapperOptions,
    mapper: IncrementalMapper,
) -> None:
    logging.info("Retriangulation and Global bundle adjustment")
    # The following is equivalent to mapper.iterative_global_refinement(...)
    custom_bundle_adjustment.iterative_global_refinement(
        mapper,
        options.ba_global_max_refinements,
        options.ba_global_max_refinement_change,
        mapper_options,
        options.get_global_bundle_adjustment(),
        options.get_triangulation(),
    )
    mapper.filter_frames(mapper_options)


def initialize_reconstruction(
    controller: IncrementalPipeline,
    mapper: IncrementalMapper,
    mapper_options: IncrementalMapperOptions,
    reconstruction: Reconstruction,
) -> IncrementalPipelineStatus:
    """Equivalent to IncrementalPipeline.initialize_reconstruction(...)"""
    options = controller.options
    init_pair = (options.init_image_id1, options.init_image_id2)

    # Try to find good initial pair
    if not options.is_initial_pair_provided():
        logging.info("Finding good initial image pair")
        ret = mapper.find_initial_image_pair(mapper_options, *init_pair)
        if ret is None:
            logging.info("No good initial image pair found.")
            return IncrementalPipelineStatus.NO_INITIAL_PAIR
        init_pair, init_cam2_from_cam1 = ret
    else:
        if not all(reconstruction.exists_image(i) for i in init_pair):
            logging.info(f"=> Initial image pair {init_pair} does not exist.")
            return IncrementalPipelineStatus.NO_INITIAL_PAIR
        maybe_init_cam2_from_cam1 = mapper.estimate_initial_two_view_geometry(
            mapper_options, *init_pair
        )
        if maybe_init_cam2_from_cam1 is None:
            logging.info("Provided pair is unsuitable for initialization")
            return IncrementalPipelineStatus.BAD_INITIAL_PAIR
        init_cam2_from_cam1 = maybe_init_cam2_from_cam1
    logging.info(
        f"Registering initial image pair #{init_pair[0]} and #{init_pair[1]}"
    )
    mapper.register_initial_image_pair(
        mapper_options, *init_pair, init_cam2_from_cam1
    )

    tri_options = options.get_triangulation()
    tri_options.min_angle = mapper_options.init_min_tri_angle
    for image_id in init_pair:
        image = reconstruction.images[image_id]
        assert image.frame is not None
        for data_id in image.frame.image_ids:
            mapper.triangulate_image(tri_options, data_id.id)

    logging.info("Global bundle adjustment")
    # The following is equivalent to: mapper.adjust_global_bundle(...)
    custom_bundle_adjustment.adjust_global_bundle(
        mapper, mapper_options, options.get_global_bundle_adjustment()
    )
    reconstruction.normalize()
    mapper.filter_points(mapper_options)
    mapper.filter_frames(mapper_options)

    # Initial image pair failed to register
    if (
        reconstruction.num_reg_frames() == 0
        or reconstruction.num_points3D() == 0
    ):
        return IncrementalPipelineStatus.BAD_INITIAL_PAIR
    if options.extract_colors:
        reconstruction.extract_colors_for_all_images(options.image_path)
    return IncrementalPipelineStatus.SUCCESS


def reconstruct_sub_model(
    controller: IncrementalPipeline,
    mapper: IncrementalMapper,
    mapper_options: IncrementalMapperOptions,
    reconstruction: Reconstruction,
) -> IncrementalPipelineStatus:
    """Equivalent to IncrementalPipeline.reconstruct_sub_model(...)"""
    mapper.begin_reconstruction(reconstruction)

    if has_unknown_sensor_from_rig(reconstruction):
        return IncrementalPipelineStatus.UNKNOWN_SENSOR_FROM_RIG

    if reconstruction.num_reg_frames() == 0:
        init_status = initialize_reconstruction(
            controller, mapper, mapper_options, reconstruction
        )
        if init_status != IncrementalPipelineStatus.SUCCESS:
            return init_status
    controller.callback(
        IncrementalPipelineCallback.INITIAL_IMAGE_PAIR_REG_CALLBACK
    )

    options = controller.options

    structure_less_flags = []
    if options.structure_less_registration_only:
        structure_less_flags = [True]
    else:
        if options.structure_less_registration_fallback:
            structure_less_flags = [False, True]
        else:
            structure_less_flags = [False]

    snapshot_prev_num_reg_frames = reconstruction.num_reg_frames()
    ba_prev_num_reg_frames = reconstruction.num_reg_frames()
    ba_prev_num_points = reconstruction.num_points3D()
    reg_next_success, prev_reg_next_success = True, True
    while True:
        if not (reg_next_success or prev_reg_next_success):
            break
        if controller.check_reached_max_runtime():
            break
        prev_reg_next_success = reg_next_success
        reg_next_success = False
        next_image_id = None
        for structure_less in structure_less_flags:
            next_images = mapper.find_next_images(
                mapper_options, structure_less=structure_less
            )
            for reg_trial, next_image_id in enumerate(next_images):
                logging.info(
                    f"Registering image #{next_image_id} "
                    f"(num_reg_frames={reconstruction.num_reg_frames()})"
                )
                if structure_less:
                    logging.info(
                        "Registering image with structure-less fallback"
                    )
                    num_vis = (
                        mapper.observation_manager.num_visible_correspondences(
                            next_image_id
                        )
                    )
                    num_corrs = mapper.observation_manager.num_correspondences(
                        next_image_id
                    )
                    logging.info(
                        f"=> Image sees {num_vis} / {num_corrs} correspondences"
                    )
                    reg_next_success = (
                        mapper.register_next_structure_less_image(
                            mapper_options, next_image_id
                        )
                    )
                else:
                    num_vis = mapper.observation_manager.num_visible_points3D(
                        next_image_id
                    )
                    num_obs = mapper.observation_manager.num_observations(
                        next_image_id
                    )
                    logging.info(f"=> Image sees {num_vis} / {num_obs} points")
                    reg_next_success = mapper.register_next_image(
                        mapper_options, next_image_id
                    )
                if reg_next_success:
                    break
                else:
                    logging.info("=> Could not register, trying another image.")
                # If initial pair fails to continue for some time,
                # abort and try different initial pair.
                kMinNumInitialRegTrials = 30
                if (
                    reg_trial >= kMinNumInitialRegTrials
                    and reconstruction.num_reg_images() < options.min_model_size
                ):
                    break
            if reg_next_success:
                break
        if reg_next_success and next_image_id is not None:
            image = reconstruction.images[next_image_id]
            assert image.frame is not None
            for data_id in image.frame.image_ids:
                mapper.triangulate_image(
                    options.get_triangulation(), data_id.id
                )
            # This is equivalent to mapper.iterative_local_refinement(...)
            custom_bundle_adjustment.iterative_local_refinement(
                mapper,
                options.ba_local_max_refinements,
                options.ba_local_max_refinement_change,
                mapper_options,
                options.get_local_bundle_adjustment(),
                options.get_triangulation(),
                next_image_id,
            )
            if controller.check_run_global_refinement(
                reconstruction, ba_prev_num_reg_frames, ba_prev_num_points
            ):
                iterative_global_refinement(options, mapper_options, mapper)
                ba_prev_num_points = reconstruction.num_points3D()
                ba_prev_num_reg_frames = reconstruction.num_reg_frames()
            if options.extract_colors:
                for data_id in image.frame.image_ids:
                    if not reconstruction.extract_colors_for_image(
                        data_id.id, options.image_path
                    ):
                        logging.warning(
                            f"Could not read image "
                            f"{reconstruction.images[data_id.id].name} "
                            f"at path {options.image_path}"
                        )
            if (
                options.snapshot_frames_freq > 0
                and reconstruction.num_reg_frames()
                >= options.snapshot_frames_freq + snapshot_prev_num_reg_frames
            ):
                snapshot_prev_num_reg_frames = reconstruction.num_reg_frames()
                write_snapshot(reconstruction, Path(options.snapshot_path))
            controller.callback(
                IncrementalPipelineCallback.NEXT_IMAGE_REG_CALLBACK
            )
        if mapper.num_shared_reg_images() >= int(options.max_model_overlap):
            break
        if (not reg_next_success) and prev_reg_next_success:
            iterative_global_refinement(options, mapper_options, mapper)

    if controller.check_reached_max_runtime():
        return pycolmap.IncrementalPipelineStatus.INTERRUPTED

    # Only run final global BA, if last incremental BA was not global
    if (
        reconstruction.num_reg_frames() > 0
        and reconstruction.num_reg_frames() != ba_prev_num_reg_frames
        and reconstruction.num_points3D() != ba_prev_num_points
    ):
        iterative_global_refinement(options, mapper_options, mapper)
    return IncrementalPipelineStatus.SUCCESS


def reconstruct(
    controller: IncrementalPipeline,
    mapper: IncrementalMapper,
    mapper_options: IncrementalMapperOptions,
    continue_reconstruction: bool,
) -> IncrementalPipelineStatus:
    """Equivalent to IncrementalPipeline.reconstruct(...)"""
    options = controller.options

    database_cache = controller.database_cache
    reconstruction_manager = controller.reconstruction_manager

    for num_trials in range(options.init_num_trials):
        if controller.check_reached_max_runtime():
            break
        if not continue_reconstruction or num_trials > 0:
            reconstruction_idx = reconstruction_manager.add()
        else:
            reconstruction_idx = 0

        reconstruction = reconstruction_manager.get(reconstruction_idx)
        status = reconstruct_sub_model(
            controller, mapper, mapper_options, reconstruction
        )
        if status == IncrementalPipelineStatus.INTERRUPTED:
            reconstruction.update_point_3d_errors()
            logging.info("Keeping reconstruction due to interrupt")
            mapper.end_reconstruction(False)
            pycolmap.align_reconstruction_to_orig_rig_scales(
                database_cache.rigs, reconstruction
            )
        elif status == IncrementalPipelineStatus.UNKNOWN_SENSOR_FROM_RIG:
            logging.error(
                "Discarding reconstruction due to unknown sensor_from_rig "
                "poses. Either explicitly define the poses by configuring the "
                "rigs or first run reconstruction without configured rigs and "
                "then derive the poses from the initial reconstruction for a "
                "subsequent reconstruction with rig constraints. See "
                "documentation for detailed instructions."
            )
            mapper.end_reconstruction(True)
            reconstruction_manager.delete(reconstruction_idx)
            return IncrementalPipelineStatus.STOP
        elif status == IncrementalPipelineStatus.BAD_INITIAL_PAIR:
            logging.info("Disacarding reconstruction due to bad initial pair")
            mapper.end_reconstruction(True)
            reconstruction_manager.delete(reconstruction_idx)
        elif status == IncrementalPipelineStatus.NO_INITIAL_PAIR:
            logging.info("Disacarding reconstruction due to no initial pair")
            mapper.end_reconstruction(True)
            reconstruction_manager.delete(reconstruction_idx)
            return IncrementalPipelineStatus.CONTINUE
        elif status == IncrementalPipelineStatus.SUCCESS:
            num_reg_images = reconstruction.num_reg_images()
            total_num_reg_images = mapper.num_total_reg_images()
            if (
                options.multiple_models
                and reconstruction_manager.size() > 1
                and num_reg_images < options.min_model_size
            ) or num_reg_images == 0:
                logging.info(
                    "Discarding reconstruction due to insufficient size"
                )
                mapper.end_reconstruction(True)
                reconstruction_manager.delete(reconstruction_idx)
            else:
                reconstruction.update_point_3d_errors()
                logging.info("Keeping successful reconstruction")
                mapper.end_reconstruction(False)
                pycolmap.align_reconstruction_to_orig_rig_scales(
                    database_cache.rigs, reconstruction
                )
            controller.callback(
                IncrementalPipelineCallback.LAST_IMAGE_REG_CALLBACK
            )
            if (
                not options.multiple_models
                or reconstruction_manager.size() >= options.max_num_models
                or total_num_reg_images >= database_cache.num_images() - 1
            ):
                return IncrementalPipelineStatus.STOP
        else:
            logging.fatal(f"Unknown reconstruction status: {status}")

    return IncrementalPipelineStatus.CONTINUE


def main_incremental_mapper(controller: IncrementalPipeline) -> None:
    """Equivalent to IncrementalPipeline.run()"""
    timer = pycolmap.Timer()
    timer.start()

    database_cache = controller.database_cache

    if database_cache.num_images() == 0:
        logging.warning("No images with matches found in the database")
        return

    if (
        controller.options.use_prior_position
        and database_cache.num_pose_priors() == 0
    ):
        logging.warning("No pose priors")
        return

    reconstruction_manager = controller.reconstruction_manager

    continue_reconstruction = reconstruction_manager.size() > 0
    if reconstruction_manager.size() > 1:
        logging.fatal(
            "Can only resume from a single reconstruction, "
            "but multiple are given"
        )

    num_images = database_cache.num_images()
    mapper = IncrementalMapper(database_cache)
    mapper_options = controller.options.get_mapper()
    if (
        reconstruct(controller, mapper, mapper_options, continue_reconstruction)
        == IncrementalPipelineStatus.STOP
    ):
        return

    def should_stop():
        return (
            mapper.num_total_reg_images() == num_images
            or controller.check_reached_max_runtime()
        )

    for _ in range(2):  # number of relaxations
        if should_stop():
            break

        logging.info("=> Relaxing the initialization constraints")
        mapper_options.init_min_num_inliers = int(
            mapper_options.init_min_num_inliers / 2
        )
        mapper.reset_initialization_stats()
        if (
            reconstruct(
                controller,
                mapper,
                mapper_options,
                continue_reconstruction=False,
            )
            == IncrementalPipelineStatus.STOP
        ):
            return

        if should_stop():
            break

        logging.info("=> Relaxing the initialization constraints")
        mapper_options.init_min_tri_angle /= 2
        mapper.reset_initialization_stats()
        if (
            reconstruct(
                controller,
                mapper,
                mapper_options,
                continue_reconstruction=False,
            )
            == IncrementalPipelineStatus.STOP
        ):
            return
    timer.print_minutes()


def main(
    database_path: Path,
    image_path: Path,
    output_path: Path,
    options: IncrementalPipelineOptions | None = None,
    input_path: Path | None = None,
) -> dict[int, Reconstruction]:
    if options is None:
        options = IncrementalPipelineOptions()
    options.image_path = image_path
    if not database_path.exists():
        logging.fatal(f"Database path does not exist: {database_path}")
    if not image_path.exists():
        logging.fatal(f"Image path does not exist: {image_path}")
    output_path.mkdir(exist_ok=True, parents=True)
    reconstruction_manager = ReconstructionManager()
    if input_path:
        reconstruction_manager.read(input_path)

    with pycolmap.Database.open(database_path) as database:
        mapper = IncrementalPipeline(options, database, reconstruction_manager)
        num_images = database.num_images()
        with enlighten.Manager() as manager:
            with manager.counter(
                total=num_images, desc="Images registered:"
            ) as pbar:
                pbar.update(0, force=True)
                mapper.add_callback(
                    IncrementalPipelineCallback.INITIAL_IMAGE_PAIR_REG_CALLBACK,
                    lambda: pbar.update(2),
                )
                mapper.add_callback(
                    IncrementalPipelineCallback.NEXT_IMAGE_REG_CALLBACK,
                    lambda: pbar.update(1),
                )
                main_incremental_mapper(mapper)

    # write and output
    reconstruction_manager.write(output_path)
    reconstructions = {}
    for i in range(reconstruction_manager.size()):
        reconstructions[i] = reconstruction_manager.get(i)
    return reconstructions


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--database_path", required=True)
    parser.add_argument("--image_path", required=True)
    parser.add_argument("--input_path", default=None)
    parser.add_argument("--output_path", required=True)
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    main(
        database_path=Path(args.database_path),
        image_path=Path(args.image_path),
        input_path=Path(args.input_path) if args.input_path else None,
        output_path=Path(args.output_path),
    )



"""
Convert legacy rotation averaging file formats to a COLMAP database.

This script provides a migration path from deprecated file-based input
to database-based input for the rotation averaging CLI.

Legacy file formats:
- Relative poses: IMAGE_NAME_1 IMAGE_NAME_2 QW QX QY QZ TX TY TZ
- Gravity priors: IMAGE_NAME GX GY GZ (optional)

After conversion, use the database with:
    colmap rotation_averager --database_path database.db --output_path output/

Usage:
    python legacy_rotation_averaging_io.py \
        --relpose_path relative_poses.txt \
        --database_path database.db \
        [--gravity_path gravity_priors.txt]
"""

import argparse
from dataclasses import dataclass
from pathlib import Path

import numpy as np
from numpy.typing import NDArray

import pycolmap
from pycolmap import logging


@dataclass
class RelativePose:
    """Relative pose between two images."""

    image_name1: str
    image_name2: str
    cam2_from_cam1: pycolmap.Rigid3d


@dataclass
class GravityPrior:
    """Gravity prior for an image."""

    image_name: str
    gravity: NDArray[np.float64]


def read_relative_poses(file_path: Path | str) -> list[RelativePose]:
    """Read relative poses from a file.

    Format: IMAGE_NAME_1 IMAGE_NAME_2 QW QX QY QZ TX TY TZ

    Args:
        file_path: Path to the relative poses file.

    Returns:
        List of RelativePose objects.
    """
    file_path = Path(file_path)
    relative_poses = []

    with open(file_path) as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            parts = line.split()
            if len(parts) < 9:
                continue

            name1, name2 = parts[0], parts[1]
            qw, qx, qy, qz = map(float, parts[2:6])
            tx, ty, tz = map(float, parts[6:9])

            # pycolmap.Rotation3d takes quaternion in xyzw format
            rotation = pycolmap.Rotation3d(np.array([qx, qy, qz, qw]))
            translation = np.array([tx, ty, tz])
            cam2_from_cam1 = pycolmap.Rigid3d(rotation, translation)

            relative_poses.append(
                RelativePose(
                    image_name1=name1,
                    image_name2=name2,
                    cam2_from_cam1=cam2_from_cam1,
                )
            )

    return relative_poses


def read_gravity_priors(file_path: Path | str) -> list[GravityPrior]:
    """Read gravity priors from a file.

    Format: IMAGE_NAME GX GY GZ

    Args:
        file_path: Path to the gravity priors file.

    Returns:
        List of GravityPrior objects.
    """
    file_path = Path(file_path)
    gravity_priors = []

    with open(file_path) as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            parts = line.split()
            if len(parts) < 4:
                continue

            name = parts[0]
            gx, gy, gz = map(float, parts[1:4])

            gravity_priors.append(
                GravityPrior(
                    image_name=name,
                    gravity=np.array([gx, gy, gz], dtype=np.float64),
                )
            )

    return gravity_priors


def get_image_names_from_relative_poses(
    relative_poses: list[RelativePose],
) -> dict[str, int]:
    """Extract unique image names from relative poses and assign IDs.

    Args:
        relative_poses: List of RelativePose objects.

    Returns:
        Dictionary mapping image names to IDs.
    """
    image_names: dict[str, int] = {}
    next_id = 1  # COLMAP uses 1-based IDs

    for pose in relative_poses:
        for name in [pose.image_name1, pose.image_name2]:
            if name not in image_names:
                image_names[name] = next_id
                next_id += 1

    return image_names


def create_database_from_relative_poses(
    database_path: Path,
    relative_poses: list[RelativePose],
    gravity_priors: list[GravityPrior] | None = None,
) -> dict[str, int]:
    """Create a COLMAP database from relative poses and gravity priors.

    Args:
        database_path: Path to the output database.
        relative_poses: List of relative poses between image pairs.
        gravity_priors: Optional list of gravity priors for images.

    Returns:
        Dictionary mapping image names to their database image IDs.
    """
    image_name_to_id = get_image_names_from_relative_poses(relative_poses)

    if database_path.exists():
        database_path.unlink()

    # Use Reconstruction to create cameras, rigs, frames, and images
    # with the trivial rig/frame helper methods
    reconstruction = pycolmap.Reconstruction()
    for image_name, image_id in image_name_to_id.items():
        camera_id = image_id

        # Create camera with trivial rig (rig_id = camera_id)
        camera = pycolmap.Camera.create(
            camera_id=camera_id,
            model=pycolmap.CameraModelId.SIMPLE_PINHOLE,
            focal_length=1.0,
            width=1,
            height=1,
        )
        reconstruction.add_camera_with_trivial_rig(camera)

        # Create image with trivial frame (frame_id = image_id)
        image = pycolmap.Image()
        image.image_id = image_id
        image.name = image_name
        image.camera_id = camera_id
        reconstruction.add_image_with_trivial_frame(image)

    # Write to database
    with pycolmap.Database.open(database_path) as db:
        for camera in reconstruction.cameras.values():
            db.write_camera(camera, use_camera_id=True)

        for rig in reconstruction.rigs.values():
            db.write_rig(rig, use_rig_id=True)

        for frame in reconstruction.frames.values():
            db.write_frame(frame, use_frame_id=True)

        for image in reconstruction.images.values():
            db.write_image(image, use_image_id=True)

        # Write two-view geometries with relative poses
        for rel_pose in relative_poses:
            id1 = image_name_to_id[rel_pose.image_name1]
            id2 = image_name_to_id[rel_pose.image_name2]

            two_view_geom = pycolmap.TwoViewGeometry()
            two_view_geom.config = (
                pycolmap.TwoViewGeometryConfiguration.CALIBRATED
            )
            two_view_geom.cam2_from_cam1 = rel_pose.cam2_from_cam1

            db.write_two_view_geometry(id1, id2, two_view_geom)

        # Write gravity priors if provided
        if gravity_priors:
            gravity_by_name = {
                gp.image_name: gp.gravity for gp in gravity_priors
            }
            for image in reconstruction.images.values():
                if image.name in gravity_by_name:
                    pose_prior = pycolmap.PosePrior()
                    pose_prior.pose_prior_id = image.image_id
                    pose_prior.corr_data_id = image.data_id
                    pose_prior.gravity = gravity_by_name[image.name]
                    db.write_pose_prior(pose_prior, use_pose_prior_id=True)

    return image_name_to_id


def main():
    parser = argparse.ArgumentParser(
        description="Convert rotation averaging files to database format"
    )
    parser.add_argument(
        "--relpose_path",
        type=Path,
        required=True,
        help="Path to relative poses file",
    )
    parser.add_argument(
        "--database_path",
        type=Path,
        required=True,
        help="Path for output database",
    )
    parser.add_argument(
        "--gravity_path",
        type=Path,
        default=None,
        help="Optional path to gravity priors file",
    )
    args = parser.parse_args()

    if not args.relpose_path.exists():
        logging.error(f"Relative poses file not found: {args.relpose_path}")
        return 1

    logging.info(f"Reading relative poses from {args.relpose_path}")
    relative_poses = read_relative_poses(args.relpose_path)
    logging.info(f"Loaded {len(relative_poses)} relative poses")

    gravity_priors = None
    if args.gravity_path is not None and args.gravity_path.exists():
        logging.info(f"Reading gravity priors from {args.gravity_path}")
        gravity_priors = read_gravity_priors(args.gravity_path)
        logging.info(f"Loaded {len(gravity_priors)} gravity priors")

    logging.info(f"Creating database at {args.database_path}")
    image_name_to_id = create_database_from_relative_poses(
        args.database_path, relative_poses, gravity_priors
    )
    logging.info(f"Database created with {len(image_name_to_id)} images")

    logging.info(
        f"Conversion complete. Use the database with:\n"
        f"  colmap rotation_averager --database_path {args.database_path} "
        f"--output_path <output_path>"
    )
    return 0


if __name__ == "__main__":
    exit(main())



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from .utils import Dataset, SceneInfo


class DatasetETH3D(Dataset):
    @property
    def position_accuracy_gt(self):
        return 0.001

    def list_scenes(self):
        scene_infos = []
        for category_path in (self.data_path / "eth3d").iterdir():
            if not category_path.is_dir() or (
                self.categories and category_path.name not in self.categories
            ):
                continue

            category = category_path.name

            for scene_path in sorted(category_path.iterdir()):
                if not scene_path.is_dir():
                    continue

                scene = scene_path.name
                if self.scenes and scene not in self.scenes:
                    continue

                workspace_path = (
                    self.run_path / self.run_name / "eth3d" / category / scene
                )
                image_path = scene_path / "images"
                sparse_gt_path = list(
                    scene_path.glob("*_calibration_undistorted")
                )[0]

                colmap_extra_args = []
                if category == "dslr":
                    colmap_extra_args.extend(["--data_type", "individual"])
                elif category == "rig":
                    colmap_extra_args.extend(["--data_type", "video"])

                scene_info = SceneInfo(
                    dataset="eth3d",
                    category=category,
                    scene=scene,
                    workspace_path=workspace_path,
                    image_path=image_path,
                    sparse_gt_path=sparse_gt_path,
                    has_camera_priors=True,
                    colmap_extra_args=colmap_extra_args,
                )

                scene_infos.append(scene_info)

        return scene_infos

    def prepare_scene(self, scene_info):
        # Nothing to prepare for ETH3D.
        pass



"""
An example for running incremental SfM on images with the pycolmap interface.
"""

import shutil
import urllib.request
import zipfile
from pathlib import Path

import enlighten

import pycolmap
from pycolmap import logging


def incremental_mapping_with_pbar(
    database_path: Path, image_path: Path, sfm_path: Path
) -> dict[int, pycolmap.Reconstruction]:
    with pycolmap.Database.open(database_path) as database:
        num_images = database.num_images()
    with enlighten.Manager() as manager:
        with manager.counter(
            total=num_images, desc="Images registered:"
        ) as pbar:
            pbar.update(0, force=True)
            reconstructions = pycolmap.incremental_mapping(
                database_path,
                image_path,
                sfm_path,
                initial_image_pair_callback=lambda: pbar.update(2),
                next_image_callback=lambda: pbar.update(1),
            )
    return reconstructions


def run() -> None:
    output_path = Path("example/")
    image_path = output_path / "Fountain/images"
    database_path = output_path / "database.db"
    sfm_path = output_path / "sfm"

    output_path.mkdir(exist_ok=True)
    # The log filename is postfixed with the execution timestamp.
    logging.set_log_destination(logging.INFO, output_path / "INFO.log.")

    data_url = "https://cvg-data.inf.ethz.ch/local-feature-evaluation-schoenberger2017/Strecha-Fountain.zip"
    if not image_path.exists():
        logging.info("Downloading the data.")
        zip_path = output_path / "data.zip"
        urllib.request.urlretrieve(data_url, zip_path)
        with zipfile.ZipFile(zip_path, "r") as fid:
            fid.extractall(output_path)
        logging.info(f"Data extracted to {output_path}.")

    if database_path.exists():
        database_path.unlink()
    pycolmap.set_random_seed(0)
    pycolmap.extract_features(database_path, image_path)
    pycolmap.match_exhaustive(database_path)

    if sfm_path.exists():
        shutil.rmtree(sfm_path)
    sfm_path.mkdir(exist_ok=True)

    recs = incremental_mapping_with_pbar(database_path, image_path, sfm_path)
    # alternatively, use:
    # import custom_incremental_pipeline
    # recs = custom_incremental_pipeline.main(
    #     database_path, image_path, sfm_path
    # )
    for idx, rec in recs.items():
        logging.info(f"#{idx} {rec.summary()}")


if __name__ == "__main__":
    run()



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

import pycolmap

from .utils import Dataset, SceneInfo


class _DatasetIMC(Dataset):
    @property
    def position_accuracy_gt(self):
        return 0.02

    def list_scenes(self):
        folder_name = f"imc{self.year}"

        scene_infos = []
        for category_path in Path(
            self.data_path / f"{folder_name}/train"
        ).iterdir():
            if not category_path.is_dir() or (
                self.categories and category_path.name not in self.categories
            ):
                continue

            category = category_path.name

            for scene_path in category_path.iterdir():
                if not scene_path.is_dir():
                    continue

                scene = scene_path.name
                if self.scenes and scene not in self.scenes:
                    continue

                sfm_path = scene_path / "sfm"
                if not sfm_path.exists():
                    pycolmap.logging.warning(
                        f"Skipping dataset=IMC{self.year}, "
                        f"category={category}, scene={scene}, "
                        "because the GT reconstruction is missing"
                    )
                    continue

                workspace_path = (
                    self.run_path
                    / self.run_name
                    / folder_name
                    / category
                    / scene
                )
                image_path = scene_path / "images"
                sparse_gt_path = scene_path / "sparse_gt"

                scene_info = SceneInfo(
                    dataset=f"IMC{self.year}",
                    category=category,
                    scene=scene,
                    workspace_path=workspace_path,
                    image_path=image_path,
                    sparse_gt_path=sparse_gt_path,
                    has_camera_priors=False,
                    colmap_extra_args=None,
                )

                scene_infos.append(scene_info)

        return scene_infos

    def prepare_scene(self, scene_info):
        if scene_info.sparse_gt_path.exists():
            return

        scene_path = scene_info.image_path.parent
        sfm_path = scene_path / "sfm"

        sparse_sfm = pycolmap.Reconstruction(sfm_path)
        sparse_gt = pycolmap.Reconstruction()

        train_image_ids = set()
        train_image_names = set(
            image.name for image in scene_info.image_path.iterdir()
        )
        for image in sparse_sfm.images.values():
            if image.name in train_image_names:
                train_image_ids.add(image.image_id)
        for camera in sparse_sfm.cameras.values():
            sparse_gt.add_camera(camera)
        for rig in sparse_sfm.rigs.values():
            sparse_gt.add_rig(rig)
        for frame in sparse_sfm.frames.values():
            has_train_image = False
            for data_id in frame.image_ids:
                if data_id.id in train_image_ids:
                    has_train_image = True
                    break
            if has_train_image:
                frame.reset_rig_ptr()
                sparse_gt.add_frame(frame)
        for image in sparse_sfm.images.values():
            if image.image_id not in train_image_ids:
                continue
            if image.camera_id not in sparse_gt.cameras:
                sparse_gt.add_camera(image.camera)
            image.reset_camera_ptr()
            image.reset_frame_ptr()
            sparse_gt.add_image(image)

        scene_info.sparse_gt_path.mkdir(exist_ok=True)
        sparse_gt.write(scene_info.sparse_gt_path)


class DatasetIMC2023(_DatasetIMC):
    def __init__(
        self,
        data_path: Path,
        categories: list[str],
        scenes: list[Path],
        run_path: Path,
        run_name: str,
    ):
        super().__init__(
            data_path=data_path,
            categories=categories,
            scenes=scenes,
            run_path=run_path,
            run_name=run_name,
        )
        self.year = 2023


class DatasetIMC2024(_DatasetIMC):
    def __init__(
        self,
        data_path: Path,
        categories: list[str],
        scenes: list[Path],
        run_path: Path,
        run_name: str,
    ):
        super().__init__(
            data_path=data_path,
            categories=categories,
            scenes=scenes,
            run_path=run_path,
            run_name=run_name,
        )
        self.year = 2024



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import collections
import copy
import dataclasses
import datetime
import functools
import multiprocessing
import shutil
import subprocess
from abc import ABC, abstractmethod
from collections.abc import Callable
from pathlib import Path

import numpy as np
import numpy.typing as npt

import pycolmap


@dataclasses.dataclass(kw_only=True)
class SceneInfo:
    # Dataset name.
    dataset: str
    # Category name.
    category: str
    # Scene name.
    scene: str
    # Path to the workspace directory in the run directory.
    workspace_path: Path
    # Path to the input images.
    image_path: Path
    # Path to the ground-truth sparse reconstruction.
    sparse_gt_path: Path
    # Whether the dataset has camera priors.
    has_camera_priors: bool
    # Additional arguments for the COLMAP reconstruction command.
    colmap_extra_args: list[str]


@dataclasses.dataclass(kw_only=True)
class SceneResult:
    # Scene information for which the result was computed.
    scene_info: SceneInfo
    # Flat list of errors.
    errors: npt.NDArray[np.floating]
    # Number of images in the scene.
    num_images: int
    # Number of registered images in the scene (over all components).
    num_reg_images: int
    # Number of components in the scene.
    num_components: int
    # Number of images in the largest component.
    largest_component: int


@dataclasses.dataclass(kw_only=True)
class Metrics:
    # Recall at specified error thresholds.
    recalls: npt.NDArray[np.floating]
    # Area under the curve (AUC) scores at specified error thresholds.
    aucs: npt.NDArray[np.floating]
    error_thresholds: npt.NDArray[np.floating]
    error_type: str
    # Number of images in the scene.
    num_images: int
    # Number of registered images in the scene (over all components).
    num_reg_images: int
    # Number of components in the scene.
    num_components: int
    # Number of images in the largest component.
    largest_component: int


MetricsByScene = dict[str, Metrics]
MetricsByCatByScene = dict[str, MetricsByScene]
MetricsByDatasetByCatByScene = dict[str, MetricsByCatByScene]


class Dataset(ABC):
    def __init__(
        self,
        data_path: Path,
        categories: list[str],
        scenes: list[Path],
        run_path: Path,
        run_name: str,
    ):
        self.data_path = data_path
        self.categories = categories
        self.scenes = scenes
        self.run_path = run_path
        self.run_name = run_name

    @property
    @abstractmethod
    def position_accuracy_gt(self) -> float:
        """Ground-truth position accuracy in meters."""
        pass

    @abstractmethod
    def list_scenes(self) -> list[SceneInfo]:
        """List all scenes to evaluate."""
        pass

    @abstractmethod
    def prepare_scene(self, scene_info: SceneInfo) -> None:
        """Prepare the scene for reconstruction."""
        pass


def parse_args() -> argparse.Namespace:
    datetime_str = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_path", default=Path(__file__).parent.parent / "data", type=Path
    )
    parser.add_argument(
        "--datasets",
        nargs="+",
        default=["eth3d", "blended-mvs", "imc2023", "imc2024"],
    )
    parser.add_argument(
        "--categories",
        nargs="+",
        default=[],
        help="Categories to evaluate, if empty all categories are evaluated.",
    )
    parser.add_argument(
        "--scenes",
        nargs="+",
        default=[],
        help="Scenes to evaluate, if empty all scenes are evaluated.",
    )
    parser.add_argument(
        "--run_path", default=Path(__file__).parent.parent / "runs", type=Path
    )
    parser.add_argument("--run_name", default=datetime_str)
    parser.add_argument("--report_name", default=f"report-{datetime_str}")
    parser.add_argument(
        "--overwrite_database", default=False, action="store_true"
    )
    parser.add_argument(
        "--overwrite_matches", default=False, action="store_true"
    )
    parser.add_argument(
        "--overwrite_reconstruction", default=False, action="store_true"
    )
    parser.add_argument(
        "--overwrite_alignment", default=False, action="store_true"
    )
    parser.add_argument("--colmap_path", required=True)
    parser.add_argument("--use_gpu", default=True, action="store_true")
    parser.add_argument("--use_cpu", dest="use_gpu", action="store_false")
    parser.add_argument(
        "--parallelism",
        type=int,
        default=multiprocessing.cpu_count(),
        help="Number of processes for parallel reconstruction.",
    )
    parser.add_argument(
        "--feature",
        default="sift",
        choices=["sift", "aliked"],
    )
    parser.add_argument(
        "--mapper",
        default="incremental",
        choices=["incremental", "hierarchical", "global"],
    )
    parser.add_argument(
        "--quality", default="high", choices=["low", "medium", "high"]
    )
    parser.add_argument(
        "--uncalibrated",
        default=False,
        action="store_true",
        help="Whether to evaluate the setting of uncalibrated input cameras, "
        "even if normal setting for the dataset contains calibrated inputs. "
        "This is useful for evaluating the performance of self-calibration.",
    )
    parser.add_argument(
        "--error_type",
        default="relative_auc",
        choices=[
            "relative_auc",
            "absolute_auc",
            "relative_recall",
            "absolute_recall",
        ],
        help="Whether to evaluate relative pairwise pose errors in angular "
        "distance or absolute pose errors through GT alignment.",
    )
    parser.add_argument(
        "--rel_error_thresholds",
        type=float,
        nargs="+",
        default=[0.5, 1, 5, 10],
        help="Evaluation thresholds in degrees.",
    )
    parser.add_argument(
        "--abs_error_thresholds",
        type=float,
        nargs="+",
        default=[0.02, 0.05, 0.2, 0.5],
        help="Evaluation thresholds in meters.",
    )
    args = parser.parse_args()
    args.colmap_path = Path(args.colmap_path).resolve()
    if args.overwrite_database:
        pycolmap.logging.info(
            "Overwriting database also overwrites reconstruction"
        )
        args.overwrite_reconstruction = True
    if args.overwrite_matches:
        pycolmap.logging.info(
            "Overwriting matches also overwrites reconstruction"
        )
        args.overwrite_reconstruction = True
    if args.overwrite_reconstruction:
        pycolmap.logging.info(
            "Overwriting reconstruction also overwrites alignment"
        )
        args.overwrite_alignment = True
    return args


def set_camera_priors(
    database_path: Path, camera_priors_sparse_gt: pycolmap.Reconstruction
) -> None:
    pycolmap.logging.info("Setting prior cameras from GT")

    with pycolmap.Database.open(str(database_path)) as database:
        images_gt_by_name = {}
        for image_gt in camera_priors_sparse_gt.images.values():
            images_gt_by_name[image_gt.name] = image_gt

        updated_camera_ids = set()
        for image in database.read_all_images():
            if image.name not in images_gt_by_name:
                pycolmap.logging.warning(
                    f"Not setting prior camera for image {image.name}, "
                    "because it does not exist in GT"
                )
                continue
            image_gt = images_gt_by_name[image.name]
            if image.camera_id in updated_camera_ids:
                continue
            camera_gt = camera_priors_sparse_gt.cameras[image_gt.camera_id]
            camera_gt.camera_id = image.camera_id
            camera_gt.has_prior_focal_length = True
            database.update_camera(camera_gt)
            updated_camera_ids.add(image.camera_id)


def colmap_reconstruction(
    args: argparse.Namespace,
    workspace_path: Path,
    image_path: Path,
    camera_priors_sparse_gt: pycolmap.Reconstruction | None = None,
    colmap_extra_args: list | None = None,
    num_threads: int = 1,
) -> None:
    workspace_path.mkdir(parents=True, exist_ok=True)

    database_path = workspace_path / "database.db"
    if args.overwrite_database and database_path.exists():
        database_path.unlink()

    sparse_path = workspace_path / "sparse"
    if args.overwrite_reconstruction and sparse_path.exists():
        shutil.rmtree(sparse_path)

    if sparse_path.exists():
        pycolmap.logging.info("Skipping reconstruction, as it already exists")
        return

    if args.overwrite_matches:
        subprocess.check_call(
            [
                args.colmap_path,
                "database_cleaner",
                "--database_path",
                database_path,
                "--type",
                "matches",
            ],
            cwd=workspace_path,
        )

    # TODO: Expose automatic reconstruction through pycolmap bindings instead
    # of using the command line interface. One blocker for this is that we
    # currently do not produce CUDA enabled pycolmap packages.
    colmap_args = [
        args.colmap_path,
        "automatic_reconstructor",
        "--image_path",
        image_path,
        "--workspace_path",
        workspace_path,
        "--use_gpu",
        "1" if args.use_gpu else "0",
        "--num_threads",
        str(num_threads),
        "--feature",
        args.feature,
        "--mapper",
        args.mapper,
        "--quality",
        args.quality,
    ]

    subprocess.check_call(
        colmap_args
        + (colmap_extra_args or [])
        + [
            "--extraction",
            "1",
            "--matching",
            "0",
            "--sparse",
            "0",
            "--dense",
            "0",
        ],
        cwd=workspace_path,
    )

    if camera_priors_sparse_gt is not None:
        set_camera_priors(database_path, camera_priors_sparse_gt)

    subprocess.check_call(
        colmap_args
        + (colmap_extra_args or [])
        + [
            "--extraction",
            "0",
            "--matching",
            "1",
            "--sparse",
            "0",
            "--dense",
            "0",
        ],
        cwd=workspace_path,
    )

    # Decouple matching from sparse reconstruction, because matching will
    # initialize an OpenGL context and Mac on Apple silicon tends to assign GUI
    # applications to the low efficiency cores but we want to use the
    # performance cores.
    subprocess.check_call(
        colmap_args
        + (colmap_extra_args or [])
        + [
            "--extraction",
            "0",
            "--matching",
            "0",
            "--sparse",
            "1",
            "--dense",
            "0",
        ],
        cwd=workspace_path,
    )


def colmap_alignment(
    args: argparse.Namespace,
    sparse_path: Path,
    sparse_gt_path: Path,
    sparse_aligned_path: Path,
    max_ref_model_error: float,
) -> None:
    if args.overwrite_alignment and sparse_aligned_path.exists():
        shutil.rmtree(sparse_aligned_path)
    if sparse_aligned_path.exists():
        pycolmap.logging.info("Skipping alignment, as it already exists")
        return

    if sparse_path.exists():
        sparse_aligned_path.mkdir(parents=True, exist_ok=True)
        subprocess.call(
            [
                args.colmap_path,
                "model_aligner",
                "--input_path",
                sparse_path,
                "--ref_model_path",
                sparse_gt_path,
                "--output_path",
                sparse_aligned_path,
                "--alignment_max_error",
                str(max_ref_model_error),
            ]
        )


def process_scene(
    args: argparse.Namespace,
    scene_info: SceneInfo,
    prepare_scene: Callable[[SceneInfo], None],
    position_accuracy_gt: float,
    num_threads: int,
) -> SceneResult:
    pycolmap.logging.info(
        f"Processing dataset={scene_info.dataset}, "
        f"category={scene_info.category}, "
        f"scene={scene_info.scene}"
    )

    prepare_scene(scene_info)

    sparse_gt = pycolmap.Reconstruction(str(scene_info.sparse_gt_path))

    colmap_reconstruction(
        args=args,
        workspace_path=scene_info.workspace_path,
        image_path=scene_info.image_path,
        camera_priors_sparse_gt=(
            sparse_gt
            if not args.uncalibrated and scene_info.has_camera_priors
            else None
        ),
        num_threads=num_threads,
        colmap_extra_args=scene_info.colmap_extra_args,
    )

    # Merge all sub-models into a single reconstruction. Each sub-model will be
    # "randomly" aligned to the other sub-models. We then compute the overall
    # error over the merged reconstruction. With this simple appraoch, there is
    # a small chance that the randomly aligned images in one sub-model are
    # correctly aligned with other sub-models and the error is therefore
    # underestimated. However, this is very unlikely to happen.
    sparse_merged = pycolmap.Reconstruction()
    num_components = 0
    largest_component = 0
    for sparse_path in (scene_info.workspace_path / "sparse").iterdir():
        if not sparse_path.is_dir():
            continue
        num_components += 1
        sparse = None
        if args.error_type.startswith("relative"):
            sparse = pycolmap.Reconstruction(str(sparse_path))
        elif args.error_type.startswith("absolute"):
            sparse_aligned_path = scene_info.workspace_path / "sparse_aligned"
            colmap_alignment(
                args=args,
                sparse_path=sparse_path,
                sparse_gt_path=scene_info.sparse_gt_path,
                sparse_aligned_path=sparse_aligned_path,
                max_ref_model_error=position_accuracy_gt,
            )
            if (sparse_aligned_path / "images.bin").exists():
                sparse = pycolmap.Reconstruction(str(sparse_aligned_path))
        else:
            raise ValueError(f"Invalid error type: {args.error_type}")

        if sparse is not None:
            largest_component = max(largest_component, sparse.num_images())
            for image in sparse.images.values():
                if image.image_id in sparse_merged.images:
                    continue
                if image.camera_id not in sparse_merged.cameras:
                    sparse_merged.add_camera(image.camera)
                if image.frame_id not in sparse_merged.frames:
                    if image.frame.rig_id not in sparse_merged.rigs:
                        sparse_merged.add_rig(image.frame.rig)
                    image.frame.reset_rig_ptr()
                    sparse_merged.add_frame(image.frame)
                image.reset_camera_ptr()
                image.reset_frame_ptr()
                sparse_merged.add_image(image)

    if args.error_type.startswith("relative"):
        dts, dRs = compute_rel_errors(
            sparse_gt=sparse_gt,
            sparse=sparse_merged,
            min_proj_center_dist=position_accuracy_gt,
        )
        errors = np.maximum(dts, dRs)
    elif args.error_type.startswith("absolute"):
        dts, dRs = compute_abs_errors(
            sparse_gt=sparse_gt,
            sparse=sparse_merged,
        )
        errors = dts
    else:
        raise ValueError(f"Invalid error type: {args.error_type}")

    return SceneResult(
        scene_info=scene_info,
        errors=errors,
        num_images=sparse_gt.num_images(),
        num_reg_images=sparse_merged.num_images(),
        num_components=num_components,
        largest_component=largest_component,
    )


def process_scenes(
    args: argparse.Namespace,
    scene_infos: list[SceneInfo],
    prepare_scene: Callable[[SceneInfo], None],
    position_accuracy_gt: float,
) -> MetricsByCatByScene:
    error_thresholds = get_error_thresholds(args)

    num_threads = min(
        args.parallelism, 2 * max(1, int(args.parallelism / len(scene_infos)))
    )
    with multiprocessing.Pool(processes=args.parallelism) as p:
        results = p.map(
            functools.partial(
                process_scene,
                args,
                prepare_scene=prepare_scene,
                position_accuracy_gt=position_accuracy_gt,
                num_threads=num_threads,
            ),
            scene_infos,
        )

    metrics: MetricsByCatByScene = collections.defaultdict(dict)
    errors_by_category: dict[str, list] = collections.defaultdict(list)
    total_num_images = 0
    total_num_reg_images = 0
    total_num_components = 0
    total_largest_components = 0
    num_scenes = len(results)
    for result in results:
        errors_by_category[result.scene_info.category].extend(result.errors)
        total_num_images += result.num_images
        total_num_reg_images += result.num_reg_images
        total_num_components += result.num_components
        total_largest_components += result.largest_component
        metrics[result.scene_info.category][result.scene_info.scene] = Metrics(
            aucs=compute_auc(
                result.errors,
                error_thresholds,
                min_error=position_accuracy_gt,
            ),
            recalls=compute_recall(result.errors, error_thresholds),
            error_thresholds=error_thresholds,
            error_type=args.error_type,
            num_images=result.num_images,
            num_reg_images=result.num_reg_images,
            num_components=result.num_components,
            largest_component=result.largest_component,
        )

    for category, errors in errors_by_category.items():
        errors_array = np.array(errors)
        metrics[category]["__all__"] = Metrics(
            aucs=compute_auc(
                errors_array,
                error_thresholds,
                min_error=position_accuracy_gt,
            ),
            recalls=compute_recall(errors_array, error_thresholds),
            error_thresholds=error_thresholds,
            error_type=args.error_type,
            num_images=total_num_images,
            num_reg_images=total_num_reg_images,
            num_components=total_num_components,
            largest_component=total_largest_components,
        )
        aucs, recalls = compute_avg_metrics(metrics[category])
        metrics[category]["__avg__"] = Metrics(
            aucs=aucs,
            recalls=recalls,
            error_thresholds=error_thresholds,
            error_type=args.error_type,
            num_images=int(round(total_num_images / num_scenes)),
            num_reg_images=int(round(total_num_reg_images / num_scenes)),
            num_components=int(round(total_num_components / num_scenes)),
            largest_component=int(round(total_largest_components / num_scenes)),
        )

    return metrics


def normalize_vec(
    vec: npt.NDArray[np.floating], eps: float = 1e-10
) -> npt.NDArray[np.floating]:
    return vec / max(eps, float(np.linalg.norm(vec)))


def vec_angular_dist_deg(
    vec1: npt.NDArray[np.floating], vec2: npt.NDArray[np.floating]
) -> float:
    cos_dist = np.clip(np.dot(normalize_vec(vec1), normalize_vec(vec2)), -1, 1)
    return np.rad2deg(np.acos(cos_dist))


def get_error_thresholds(args: argparse.Namespace) -> npt.NDArray[np.floating]:
    if args.error_type.startswith("relative"):
        return np.array(args.rel_error_thresholds)
    elif args.error_type.startswith("absolute"):
        return np.array(args.abs_error_thresholds)
    else:
        raise ValueError(f"Invalid error type: {args.error_type}")


def get_scores(error_type: str, metrics: Metrics) -> npt.NDArray[np.floating]:
    if error_type.endswith("auc"):
        return metrics.aucs
    elif error_type.endswith("recall"):
        return metrics.recalls
    else:
        raise ValueError(f"Invalid error type: {error_type}")


def compute_rel_errors(
    sparse_gt: pycolmap.Reconstruction,
    sparse: pycolmap.Reconstruction,
    min_proj_center_dist: float,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """Computes angular relative pose errors across all image pairs.

    Notice that this approach leads to a super-linear decrease in the AUC scores
    when multiple images fail to register. Consider that we have N images in
    total in a dataset and M images are registered in the evaluated
    reconstruction. In this case, we can compute "finite" errors for (N-M)^2
    pairs while the dataset has a total of N^2 pairs. In case of many
    unregistered images, the AUC score will drop much more than the
    (intuitively) expected (N-M) / N ratio. One could appropriately normalize by
    computing a single score per image through a suitable normalization of all
    pairwise errors per image. However, this becomes difficult when multiple
    sub-components are incorrectly stitched together in the same reconstruction
    (e.g., in the case of symmetry issues).
    """

    if sparse is None:
        pycolmap.logging.error("Reconstruction failed")
        return len(sparse_gt.images) * [np.inf], len(sparse_gt.images) * [180]

    images = {}
    for image in sparse.images.values():
        images[image.name] = image

    dts = []
    dRs = []
    for this_image_gt in sparse_gt.images.values():
        if this_image_gt.name not in images:
            for _ in range(sparse_gt.num_images() - 1):
                dts.append(np.inf)
                dRs.append(180)
            continue

        this_image = images[this_image_gt.name]

        for other_image_gt in sparse_gt.images.values():
            if this_image_gt.image_id == other_image_gt.image_id:
                continue

            if other_image_gt.name not in images:
                dts.append(np.inf)
                dRs.append(180)
                continue

            other_image = images[other_image_gt.name]

            other_from_this = (
                other_image.cam_from_world()
                * this_image.cam_from_world().inverse()
            )
            other_from_this_gt = (
                other_image_gt.cam_from_world()
                * this_image_gt.cam_from_world().inverse()
            )

            estimated_from_gt = other_from_this.inverse() * other_from_this_gt

            if (
                np.linalg.norm(other_from_this_gt.translation)
                < min_proj_center_dist
            ):
                # If the cameras almost coincide, then the angular direction
                # distance is unstable, because a small position change can
                # cause a large rotational error. In this case, we only measure
                # rotational relative pose error.
                dt = 0.0
            else:
                dt = vec_angular_dist_deg(
                    other_from_this.translation, other_from_this_gt.translation
                )

            dR = np.rad2deg(estimated_from_gt.rotation.angle())

            dts.append(dt)
            dRs.append(dR)

    return np.array(dts), np.array(dRs)


def compute_abs_errors(
    sparse_gt: pycolmap.Reconstruction, sparse: pycolmap.Reconstruction
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """Computes rotational and translational absolute pose errors.

    Assumes that the input reconstructions are aligned in the same coordinate
    system. Computes one error per ground-truth image.
    """

    dts = np.full(len(sparse_gt.images), fill_value=np.inf, dtype=np.float64)
    dRs = np.full(len(sparse_gt.images), fill_value=180, dtype=np.float64)

    if sparse is None:
        pycolmap.logging.error("Reconstruction or alignment failed")
        return dts, dRs

    images = {}
    for image in sparse.images.values():
        images[image.name] = image

    dts = np.full(len(sparse_gt.images), fill_value=np.inf, dtype=np.float64)
    dRs = np.full(len(sparse_gt.images), fill_value=180, dtype=np.float64)
    for i, image_gt in enumerate(sparse_gt.images.values()):
        if image_gt.name not in images:
            continue

        image = images[image_gt.name]

        estimated_from_gt = (
            image.cam_from_world() * image_gt.cam_from_world().inverse()
        )

        dts[i] = np.linalg.norm(estimated_from_gt.translation)
        dRs[i] = np.rad2deg(estimated_from_gt.rotation.angle())

    return dts, dRs


def compute_auc(
    errors: npt.NDArray[np.floating],
    thresholds: npt.NDArray[np.floating],
    min_error: float = 0,
) -> npt.NDArray[np.floating]:
    num_elems = len(errors)
    if len(errors) == 0:
        raise ValueError("No errors to evaluate")

    errors = np.sort(errors)
    recalls = (np.arange(num_elems) + 1) / num_elems

    if min_error > 0:
        min_index = np.searchsorted(errors, min_error, side="right")
        min_recall = min_index / num_elems
        recalls = np.r_[min_recall, min_recall, recalls[min_index:]]
        errors = np.r_[0, min_error, errors[min_index:]]
    else:
        recalls = np.r_[0, recalls]
        errors = np.r_[0, errors]

    aucs = np.zeros(len(thresholds), dtype=np.float64)
    for i, t in enumerate(thresholds):
        last_index = np.searchsorted(errors, t, side="right")
        r = np.r_[recalls[:last_index], recalls[last_index - 1]]
        e = np.r_[errors[:last_index], t]
        auc = np.trapezoid(r, x=e) / t
        aucs[i] = auc * 100

    return aucs


def compute_recall(
    errors: npt.NDArray[np.floating],
    thresholds: npt.NDArray[np.floating],
    min_error: float = 0,
) -> npt.NDArray[np.floating]:
    num_elems = len(errors)
    if num_elems == 0:
        raise ValueError("No errors to evaluate")

    recalls = np.zeros(len(thresholds), dtype=np.float64)
    for i, t in enumerate(thresholds):
        recalls[i] = 100 * np.sum(errors <= t) / num_elems

    return recalls


def compute_avg_metrics(
    scene_metrics: MetricsByScene,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    auc_sum = None
    recall_sum = None
    num_scenes = 0
    for scene, metrics in scene_metrics.items():
        if scene.startswith("__") and scene.endswith("__"):
            continue
        num_scenes += 1
        if auc_sum is None:
            auc_sum = copy.copy(metrics.aucs)
        if recall_sum is None:
            recall_sum = copy.copy(metrics.recalls)
        else:
            for i in range(len(auc_sum)):
                auc_sum[i] += metrics.aucs[i]
            for i in range(len(recall_sum)):
                recall_sum[i] += metrics.recalls[i]
    return np.array(auc_sum) / num_scenes, np.array(recall_sum) / num_scenes


def diff_metrics(
    metrics_a: MetricsByDatasetByCatByScene,
    metrics_b: MetricsByDatasetByCatByScene,
):
    """Computes difference between two sets of metrics.

    Raises exception if the metrics are inconsistent.
    """
    metrics_diff = copy.deepcopy(metrics_a)
    for dataset, category_metrics_a in metrics_a.items():
        if dataset not in metrics_b:
            raise ValueError(f"Dataset {dataset} not found in metrics_b")
        category_metrics_b = metrics_b[dataset]
        for category, scene_metrics_a in category_metrics_a.items():
            if category not in category_metrics_b:
                raise ValueError(f"Category {category} not found in metrics_b")
            scene_metrics_b = category_metrics_b[category]
            for scene, metrics_a_item in scene_metrics_a.items():
                if scene not in scene_metrics_b:
                    raise ValueError(f"Scene {scene} not found in metrics_b")
                metrics_b_item = scene_metrics_b[scene]
                if (
                    metrics_a_item.error_type != metrics_b_item.error_type
                    or not np.all(
                        metrics_a_item.error_thresholds
                        == metrics_b_item.error_thresholds
                    )
                ):
                    raise ValueError("Inconsistent error thresholds or types")
                metrics_diff[dataset][category][scene] = Metrics(
                    aucs=metrics_a_item.aucs - metrics_b_item.aucs,
                    recalls=metrics_a_item.recalls - metrics_b_item.recalls,
                    error_thresholds=metrics_a_item.error_thresholds,
                    error_type=metrics_a_item.error_type,
                    num_images=metrics_a_item.num_images
                    - metrics_b_item.num_images,
                    num_reg_images=metrics_a_item.num_reg_images
                    - metrics_b_item.num_reg_images,
                    num_components=metrics_a_item.num_components
                    - metrics_b_item.num_components,
                    largest_component=metrics_a_item.largest_component
                    - metrics_b_item.largest_component,
                )
    return metrics_diff


def create_result_table(
    dataset_metrics: MetricsByDatasetByCatByScene,
) -> str:
    first_metrics = next(
        iter(next(iter(next(iter(dataset_metrics.values())).values())).values())
    )

    is_auc = first_metrics.error_type.endswith("auc")
    is_relative = first_metrics.error_type.startswith("relative")
    score_type = "AUC" if is_auc else "Recall"
    score_unit = "deg" if is_relative else "cm"
    label = f"{score_type} @ X {score_unit} (%)"
    if is_relative:
        thresholds = first_metrics.error_thresholds
    else:
        thresholds = 100 * first_metrics.error_thresholds  # cm

    column = "scenes"
    size_scenes = max(
        len(column) + 2,
        max(
            len(s)
            for d in dataset_metrics.values()
            for c in d.values()
            for s in c
        ),
    )
    size_aucs = max(len(label) + 2, len(thresholds) * 7 - 1)
    size_imgs = 12
    size_comps = 12
    size_sep = size_scenes + size_aucs + size_imgs + size_comps + 3
    header = (
        f"{column:=^{size_scenes}} {label:=^{size_aucs}} "
        f"{'images':=^{size_imgs}} {'components':=^{size_comps}}"
    )
    header += "\n" + " " * (size_scenes + 1)
    header += " ".join(f"{str(t).rstrip('.'):^6}" for t in thresholds)
    header += "    reg   all  num largest"
    text = [header]
    for dataset, category_metrics in dataset_metrics.items():
        for category, scene_metrics in category_metrics.items():
            text.append(f"\n{dataset + '=' + category:=^{size_sep}}")
            for scene, metrics in scene_metrics.items():
                scores = get_scores(first_metrics.error_type, metrics)
                assert len(scores) == len(thresholds)
                row = ""
                if scene == "__avg__":
                    scene = "average"
                    row += "-" * size_sep + "\n"
                if scene == "__all__":
                    scene = "overall"
                    row += "-" * size_sep + "\n"
                row += f"{scene:<{size_scenes}} "
                row += " ".join(f"{score:>6.2f}" for score in scores)
                row += f" {metrics.num_reg_images:6d}"
                row += f"{metrics.num_images:6d}"
                row += f" {metrics.num_components:4d}"
                row += f"{metrics.largest_component:8d}"
                text.append(row)
    return "\n".join(text)



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import numpy as np
from PIL import Image

import pycolmap

from .utils import Dataset, SceneInfo


class DatasetBlendedMVS(Dataset):
    @property
    def position_accuracy_gt(self):
        return 0.001

    def list_scenes(self):
        scene_infos = []
        for category_path in (self.data_path / "blended-mvs").iterdir():
            if not category_path.is_dir() or (
                self.categories and category_path.name not in self.categories
            ):
                continue

            category = category_path.name

            for scene_path in sorted(category_path.iterdir()):
                if not scene_path.is_dir():
                    continue

                scene = scene_path.name
                if self.scenes and scene not in self.scenes:
                    continue

                workspace_path = (
                    self.run_path
                    / self.run_name
                    / "blended-mvs"
                    / category
                    / scene
                )
                image_path = scene_path / "blended_images"
                image_list_path = scene_path / "images.txt"
                with open(image_list_path, "w") as fid:
                    for filepath in sorted(image_path.iterdir()):
                        image_name = str(filepath.name)
                        if (
                            image_name.endswith(".jpg")
                            and "masked" not in image_name
                        ):
                            fid.write(image_name + "\n")

                sparse_gt_path = scene_path / "sparse_gt"
                colmap_extra_args = ["--image_list_path", image_list_path]

                scene_info = SceneInfo(
                    dataset="blended-mvs",
                    category=category,
                    scene=scene,
                    workspace_path=workspace_path,
                    image_path=image_path,
                    sparse_gt_path=sparse_gt_path,
                    has_camera_priors=True,
                    colmap_extra_args=colmap_extra_args,
                )

                scene_infos.append(scene_info)

        return scene_infos

    def prepare_scene(self, scene_info):
        if scene_info.sparse_gt_path.exists():
            return

        scene_path = scene_info.image_path.parent

        sparse_gt = pycolmap.Reconstruction()
        for i, filepath in enumerate(sorted((scene_path / "cams").iterdir())):
            filename = str(filepath.name)
            if not filename.endswith("_cam.txt"):
                continue
            image_name = filename[:-8] + ".jpg"
            width, height = Image.open(
                scene_path / "blended_images" / image_name
            ).size[:2]
            with open(filepath, encoding="ascii") as fid:
                lines = list(map(lambda b: b.strip(), fid.readlines()))
                extrinsic = np.fromstring(
                    " ".join(lines[1:4]),
                    count=12,
                    sep=" ",
                ).reshape(3, 4)
                intrinsic = np.fromstring(
                    " ".join(lines[7:10]),
                    count=9,
                    sep=" ",
                ).reshape(3, 3)
            camera = pycolmap.Camera(
                camera_id=i,
                model=pycolmap.CameraModelId.PINHOLE,
                width=width,
                height=height,
                params=intrinsic[(0, 1, 0, 1), (0, 1, 2, 2)],
            )
            rig = pycolmap.Rig(rig_id=i)
            rig.add_ref_sensor(camera.sensor_id)
            image = pycolmap.Image(
                image_id=i,
                camera_id=i,
                name=image_name,
                cam_from_world=pycolmap.Rigid3d(extrinsic),
            )
            frame = pycolmap.Frame(frame_id=i)
            frame.add_data_id(image.data_id)
            sparse_gt.add_camera(camera)
            sparse_gt.add_rig(rig)
            sparse_gt.add_image(image)
            sparse_gt.add_frame(frame)

        scene_info.sparse_gt_path.mkdir(exist_ok=True)
        sparse_gt.write(scene_info.sparse_gt_path)



import contextlib
import ctypes
import importlib
import platform
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

from .utils import import_module_symbols


def _preload_cuda_lib(module_name: str, lib_name: str):
    """Preload a single library."""
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        return
    else:
        # TODO: update the logic to handle CUDA 13,
        # using as reference https://github.com/pytorch/pytorch/pull/163661.

        # Resolve the library directory robustly
        if module_file_path := getattr(module, "__file__", None):
            lib_dir = Path(module_file_path).parent
        elif paths := getattr(module, "__path__", None):
            # Implicit namespace packages have __path__ but no __file__
            lib_dir = Path(list(paths)[0])
        else:
            return
        # Find the first file matching the pattern
        if lib_path := next((lib_dir / "lib").glob(lib_name), None):
            with contextlib.suppress(OSError):
                ctypes.CDLL(str(lib_path))


def _preload_cuda_deps():
    """Preloads CUDA dependencies from pip packages on Linux."""
    if platform.system() != "Linux":
        return
    cuda_libs = {
        "nvidia.cuda_runtime": "libcudart.so.*[0-9]",
        "nvidia.curand": "libcurand.so.*[0-9]",
    }
    for module_name, lib_glob in cuda_libs.items():
        _preload_cuda_lib(module_name, lib_glob)


_preload_cuda_deps()

try:
    from . import _core
except ImportError as e:
    raise RuntimeError(
        textwrap.dedent("""
        Cannot import the C++ backend pycolmap._core.
        Make sure that you successfully install the package with
          $ python -m pip install pycolmap/
        """)
    ) from e

# Type checkers cannot deal with dynamic manipulation of globals.
# Instead, we use the same workaround as PyTorch.
if TYPE_CHECKING:
    from ._core import *  # noqa F403

__all__ = import_module_symbols(
    globals(), _core, exclude={"cost_functions", "pyceres"}
)
__all__.extend(["__version__", "__ceres_version__"])

__version__ = _core.__version__
__ceres_version__ = _core.__ceres_version__



import argparse
from pathlib import Path

import tomlkit

# Set up command-line argument parser
parser = argparse.ArgumentParser(
    description="Modify pyproject.toml for a custom build."
)
parser.add_argument(
    "--name", required=True, help="The new package name for the wheel."
)
parser.add_argument(
    "--add-deps",
    nargs="+",
    default=[],
    help="Space-separated list of Python dependencies to add.",
)
args = parser.parse_args()

# Modify the pyproject.toml file
pyproject_path = Path("pyproject.toml")
if not pyproject_path.exists():
    raise FileNotFoundError(pyproject_path)

with open(pyproject_path, encoding="utf-8") as f:
    config = tomlkit.load(f)

config["project"]["name"] = args.name

if args.add_deps:
    if "dependencies" not in config["project"]:
        config["project"]["dependencies"] = []

    existing_deps = config["project"]["dependencies"]
    for req in args.add_deps:
        if req not in existing_deps:
            existing_deps.append(req)

with open(pyproject_path, "w", encoding="utf-8") as f:
    tomlkit.dump(config, f)



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import pickle
from pathlib import Path

from evaluation.utils import create_result_table, diff_metrics

import pycolmap


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--report_a_path", type=Path, required=True)
    parser.add_argument("--report_b_path", type=Path, required=True)
    return parser.parse_args()


def main() -> None:
    args = parse_args()

    with open(args.report_a_path, "rb") as report_file:
        metrics_a = pickle.load(report_file)
    with open(args.report_b_path, "rb") as report_file:
        metrics_b = pickle.load(report_file)

    metrics_diff = diff_metrics(metrics_a, metrics_b)

    pycolmap.logging.info("Results A:\n" + create_result_table(metrics_a))
    pycolmap.logging.info("Results B:\n" + create_result_table(metrics_b))
    pycolmap.logging.info(
        "Results A - B:\n" + create_result_table(metrics_diff)
    )


if __name__ == "__main__":
    main()



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import argparse
import datetime
import multiprocessing
import os
import time
import urllib.error
import urllib.request
import xml.etree.ElementTree as ElementTree
from typing import Final

import urllib2
import urlparse

PER_PAGE: Final[int] = 500
SORT: Final[str] = "date-posted-desc"
URL: Final[str] = (
    "https://api.flickr.com/services/rest/?method=flickr.photos.search&"
    "api_key=%s&text=%s&sort=%s&per_page=%d&page=%d&min_upload_date=%s&"
    "max_upload_date=%s&format=rest&extras=url_o,url_l,url_c,url_z,url_n"
)
MAX_PAGE_REQUESTS: Final[int] = 5
MAX_PAGE_TIMEOUT: Final[int] = 20
MAX_IMAGE_REQUESTS: Final[int] = 3
TIME_SKIP: Final[int] = 24 * 60 * 60
MAX_DATE: Final[float] = time.time()
MIN_DATE: Final[float] = MAX_DATE - TIME_SKIP


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--search_text", required=True)
    parser.add_argument("--api_key", required=True)
    parser.add_argument("--image_path", required=True)
    parser.add_argument("--num_procs", type=int, default=10)
    parser.add_argument("--max_days_without_image", type=int, default=365)
    args = parser.parse_args()
    return args


def compose_url(
    page: int, api_key: str, text: str, min_date: float, max_date: float
) -> str:
    return URL % (
        api_key,
        text,
        SORT,
        PER_PAGE,
        page,
        str(min_date),
        str(max_date),
    )


def parse_page(
    page: int, api_key: str, text: str, min_date: float, max_date: float
) -> tuple[dict[str, str], tuple[dict[str, str], ...]]:
    f: urllib2.urlopen | None = None
    for _ in range(MAX_PAGE_REQUESTS):
        try:
            f = urllib2.urlopen(
                compose_url(page, api_key, text, min_date, max_date),
                timeout=MAX_PAGE_TIMEOUT,
            )
        except TimeoutError:
            continue
        else:
            break

    if f is None:
        return {
            "pages": "0",
            "total": "0",
            "page": "0",
            "perpage": "0",
        }, tuple()

    response = f.read()
    root = ElementTree.fromstring(response)

    if root.attrib["stat"] != "ok":
        raise OSError

    metadata = root.find("photos")
    assert metadata is not None

    photos = tuple(photo.attrib for photo in root.iter("photo"))

    return metadata.attrib, photos


class PhotoDownloader:
    def __init__(self, image_path: str) -> None:
        self.image_path: str = image_path

    def __call__(self, photo: dict[str, str]) -> None:
        # Find the URL corresponding to the highest image resolution. We will
        # need this URL here to determine the image extension (typically .jpg,
        # but could be .png, .gif, etc).
        url: str | None = None
        for url_suffix in ("o", "l", "k", "h", "b", "c", "z"):
            url_attr = f"url_{url_suffix}"
            if photo.get(url_attr) is not None:
                url = photo.get(url_attr)
                break

        if url is not None:
            # Note that the following statement may fail in Python 3. urlparse
            # may need to be replaced with urllib.parse.
            url_filename = urlparse.urlparse(url).path
            image_ext = os.path.splitext(url_filename)[1]

            image_name = f"{photo['id']}_{photo['secret']}{image_ext}"
            path = os.path.join(self.image_path, image_name)
            if not os.path.exists(path):
                print(url)
                for _ in range(MAX_IMAGE_REQUESTS):
                    try:
                        urllib.request.urlretrieve(url, path)
                    except urllib.error.ContentTooShortError:
                        continue
                    else:
                        break


def main() -> None:
    args = parse_args()

    downloader = PhotoDownloader(args.image_path)
    pool = multiprocessing.Pool(processes=args.num_procs)

    num_pages = float("inf")
    page = 0

    min_date = MIN_DATE
    max_date = MAX_DATE

    days_in_row = 0

    search_text = args.search_text.replace(" ", "-")

    while num_pages > page:
        page += 1

        metadata, photos = parse_page(
            page, args.api_key, search_text, min_date, max_date
        )

        num_pages = int(metadata["pages"])

        print(78 * "=")
        print("Page:\t\t", page, "of", num_pages)
        print("Min-Date:\t", datetime.datetime.fromtimestamp(min_date))
        print("Max-Date:\t", datetime.datetime.fromtimestamp(max_date))
        print("Num-Photos:\t", len(photos))
        print(78 * "=")

        try:
            pool.map_async(downloader, photos).get(1e10)
        except KeyboardInterrupt:
            pool.close()
            pool.join()
            break

        if page >= num_pages:
            max_date -= TIME_SKIP
            min_date -= TIME_SKIP
            page = 0

        if num_pages == 0:
            days_in_row = days_in_row + 1
            num_pages = float("inf")

            print("    No images in", days_in_row, "days in a row")

            if days_in_row == args.max_days_without_image:
                break
        else:
            days_in_row = 0


if __name__ == "__main__":
    main()



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import shutil
import subprocess
import zipfile
from pathlib import Path

import py7zr
import requests

import pycolmap


def download_file(url: str, target_folder: Path) -> str:
    filename = url.split("/")[-1]
    with requests.get(url, stream=True) as req:
        req.raise_for_status()
        with open(target_folder / filename, "wb") as f:
            for chunk in req.iter_content(chunk_size=8192):
                f.write(chunk)
    return filename


def download_eth3d(data_path: Path) -> None:
    for filename, category in [
        ("multi_view_training_dslr_undistorted.7z", "dslr"),
        ("multi_view_test_dslr_undistorted.7z", "dslr"),
        ("multi_view_training_rig_undistorted.7z", "rig"),
        ("multi_view_test_rig_undistorted.7z", "rig"),
    ]:
        target_folder = data_path / category
        target_folder.mkdir(parents=True, exist_ok=True)

        pycolmap.logging.info(
            f"Downloading ETH3D category={category}, filename={filename}"
        )
        download_file("https://www.eth3d.net/data/" + filename, target_folder)

        pycolmap.logging.info(
            f"Extracting ETH3D category={category}, filename={filename}"
        )
        with py7zr.SevenZipFile(target_folder / filename, mode="r") as archive:
            archive.extractall(path=target_folder)


def download_imc2023(data_path: Path) -> None:
    data_path.mkdir(parents=True, exist_ok=True)

    pycolmap.logging.info("Downloading IMC2023")
    subprocess.check_call(
        [
            "kaggle",
            "competitions",
            "download",
            "-c",
            "image-matching-challenge-2023",
            "-p",
            str(data_path),
        ],
    )

    pycolmap.logging.info("Extracting IMC2023")
    with zipfile.ZipFile(
        data_path / "image-matching-challenge-2023.zip", mode="r"
    ) as archive:
        archive.extractall(path=data_path)


def download_imc2024(data_path: Path) -> None:
    data_path.mkdir(parents=True, exist_ok=True)

    pycolmap.logging.info("Downloading IMC2024")
    subprocess.check_call(
        [
            "kaggle",
            "competitions",
            "download",
            "-c",
            "image-matching-challenge-2024",
            "-p",
            str(data_path),
        ],
    )

    pycolmap.logging.info("Extracting IMC2024")
    with zipfile.ZipFile(
        data_path / "image-matching-challenge-2024.zip", mode="r"
    ) as archive:
        archive.extractall(path=data_path)

    # Move all scenes to the "all" category sub-folder.
    category_path = data_path / "train/all"
    category_path.mkdir(parents=True, exist_ok=True)
    for scene in (data_path / "train").iterdir():
        if str(scene).endswith("/all"):
            continue
        shutil.move(scene, data_path / category_path)


# TODO: BlendedMVS+ and BlendedMVS++.
def download_blended_mvs(data_path: Path) -> None:
    target_folder = data_path / "BlendedMVS"
    target_folder.mkdir(parents=True, exist_ok=True)

    pycolmap.logging.info("Downloading BlendedMVS")
    for filename in [
        "BlendedMVS.zip",
    ] + [f"BlendedMVS.z{i:02d}" for i in range(1, 16)]:
        download_file(
            "https://github.com/YoYo000/BlendedMVS/releases/download/v1.0.0/"
            + filename,
            target_folder,
        )

    pycolmap.logging.info("Extracting BlendedMVS")
    with zipfile.ZipFile(target_folder / "BlendedMVS.zip", mode="r") as archive:
        archive.extractall(path=data_path)


DOWNLOADERS = {
    "eth3d": download_eth3d,
    "imc2023": download_imc2023,
    "imc2024": download_imc2024,
    "blended-mvs": download_blended_mvs,
}


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_path", type=Path, default=Path(__file__).parent / "data"
    )
    parser.add_argument(
        "--datasets",
        nargs="+",
        default=DOWNLOADERS.keys(),
        choices=DOWNLOADERS.keys(),
    )
    return parser.parse_args()


def main() -> None:
    args = parse_args()

    for dataset in args.datasets:
        DOWNLOADERS[dataset](args.data_path / dataset)


if __name__ == "__main__":
    main()



# Copyright (c), ETH Zurich and UNC Chapel Hill.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of ETH Zurich and UNC Chapel Hill nor the names of
#       its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import pickle

from evaluation.blended_mvs import DatasetBlendedMVS
from evaluation.eth3d import DatasetETH3D
from evaluation.imc import DatasetIMC2023, DatasetIMC2024
from evaluation.utils import (
    Dataset,
    create_result_table,
    parse_args,
    process_scenes,
)

import pycolmap


def main() -> None:
    args = parse_args()

    datasets: dict[str, type[Dataset]] = {
        "eth3d": DatasetETH3D,
        "blended-mvs": DatasetBlendedMVS,
        "imc2023": DatasetIMC2023,
        "imc2024": DatasetIMC2024,
    }

    metrics = {}
    for dataset_name in args.datasets:
        if dataset_name not in datasets:
            pycolmap.logging.error(f"Unknown dataset: {dataset_name}")
            return

        pycolmap.logging.info(f"Evaluating dataset: {dataset_name}")

        dataset = datasets[dataset_name](
            data_path=args.data_path,
            categories=args.categories,
            scenes=args.scenes,
            run_path=args.run_path,
            run_name=args.run_name,
        )

        scene_infos = dataset.list_scenes()

        if not scene_infos:
            pycolmap.logging.warning("No scenes found")
            return

        metrics[dataset_name] = process_scenes(
            args=args,
            scene_infos=scene_infos,
            prepare_scene=dataset.prepare_scene,
            position_accuracy_gt=dataset.position_accuracy_gt,
        )

    pycolmap.logging.info("Results:\n" + create_result_table(metrics))

    report_path = args.run_path / args.run_name / (args.report_name + ".pkl")
    pycolmap.logging.info(f"Saving report to: {report_path}")
    with open(report_path, "wb") as report_file:
        pickle.dump(metrics, report_file)


if __name__ == "__main__":
    main()
